from collections import defaultdict

from .base_transformer import BaseTransformer


class AutomatTransformer(BaseTransformer):
    def wrap(self, res):
        res = {"data": res}
        result = defaultdict(set)

        for rec in res["data"]:
            obj = rec[-1]
            obj_id_prefix = obj["id"].split(":")[0]
            if obj_id_prefix in ["MONDO", "CHEBI", "DOID"]:
                result[obj_id_prefix.lower()].add(obj["id"])
            elif obj_id_prefix in ["NCBIGene", "ENSEMBL"]:
                result[obj_id_prefix.lower()].add(obj["id"].split(":", 1)[-1])
            elif obj_id_prefix == "CHEMBL.COMPOUND":
                result["chembl"].add(obj["id"].split(":", 1)[-1])
            obj = rec[0]
            obj_id_prefix = obj["id"].split(":")[0]
            if obj_id_prefix in ["MONDO", "CHEBI", "DOID"]:
                result[obj_id_prefix.lower()].add(obj["id"])
            elif obj_id_prefix in ["NCBIGene", "ENSEMBL"]:
                result[obj_id_prefix.lower()].add(obj["id"].split(":", 1)[-1])
            elif obj_id_prefix == "CHEMBL.COMPOUND":
                result["chembl"].add(obj["id"].split(":", 1)[-1])
        new_result = {}
        for k, v in result.items():
            new_result[k] = list(v)
        return {"associated_with": new_result}




from .base_transformer import BaseTransformer
from .utils import generateCurie


class BioThingsTransformer(BaseTransformer):
    def pairInputWithAPIResponse(self):
        if self.edge["query_operation"]["method"] == "post":
            res = {}
            for item in self.data["response"]:
                if "notfound" not in item:
                    _input = generateCurie(
                        self.edge["association"]["input_id"], item["query"]
                    )
                    if _input in res:
                        res[_input].append(item)
                    else:
                        res[_input] = [item]
            return res
        else:
            _input = generateCurie(
                self.edge["association"]["input_id"], self.edge["input"]
            )
            return {_input: [self.data["response"]]}



def restructure_biolink_response(json_doc):
    """ restructure API response from biolink API before extracting data

    parameters
    ----------
    json_doc: the API response from biolink API

    notes: list of prefixes used in biolink API for different semantic types

        * ANATOMY: UBERON, CL, FBbt
        * DISEASE: MONDO
        * GENE: HGNC, NCBIGene, MGI， ZFIN，FlyBase
        * PHENOTYPE: EFO, HP, MONDO
    """
    if json_doc and 'associations' in json_doc:
        for _doc in json_doc['associations']:
            # remove prefix
            if 'object' in _doc and 'id' in _doc['object']:
                object_id = _doc['object']['id']
                try:
                    prefix, value = object_id.split(':')
                    # these IDs have prefix by nature
                    if prefix in ['HGNC', 'NCBIGene', 'REACT']:
                        _doc['object'][prefix] = value
                    else:
                        _doc['object'][prefix] = object_id
                except ValueError:
                    pass
            # remove empty value
            if 'publications' not in _doc:
                continue
            if not _doc.get('publications'):
                _doc.pop('publications')
            elif not _doc['publications'][0]['id'].startswith('PMID'):
                _doc.pop('publications')
            else:
                for _item in _doc['publications']:
                    _item['id'] = _item['id'].split(':')[-1]
            if not _doc['provided_by']:
                _doc.pop('provided_by')
    return json_doc



# -*- coding: utf-8 -*-
"""
Export the output of BTE to vis.js accepted format.

.. moduleauthor:: Jiwen Xin <kevinxin@scripps.edu>


"""


def networkx_json_to_visjs(res):
    """Convert JSON output from networkx to visjs compatible version.

    :param: res: JSON output from networkx of the graph
    """
    colors = {1: 'green', 2: 'red', 3: 'rgba(255,168,7)'}
    if res:
        links = res['links']
        new_links = []
        for _link in links:
            _link['from'] = _link.pop('source')
            _link['to'] = _link.pop('target')
            _link['font'] = {'align': 'middle'}
            _link['arrows'] = 'to'
            new_links.append(_link)
        res['links'] = new_links
        new_nodes = []
        for _node in res['nodes']:
            _node['label'] = _node['identifier'] + ':' + str(_node['id'])
            _node['color'] = colors[_node['level']]
            if 'equivalent_ids' in _node:
                equ_ids = []
                for k, v in _node['equivalent_ids'].items():
                    if isinstance(v, list):
                        for _v in v:
                            equ_ids.append(k + ':' + str(_v))
                    else:
                        equ_ids.append(k + ":" + str(v))
                equ_ids = '<br>'.join(equ_ids)
                _node['equivalent_ids'] = equ_ids
            new_nodes.append(_node)
        res['nodes'] = new_nodes
    return res



import asyncio
import json
import traceback
from json import JSONDecodeError
from math import floor
from collections import Counter, defaultdict
from aiohttp import ClientSession, TCPConnector, ClientTimeout

from ..resolve_ids import asyncQuery
from .query_builder import QueryBuilder
from .api_response_transform import Transformer
from ..config_new import MAX_CONCURRENT_QUERIES_ON_SINGLE_API
from .filter import filter_response


class APIQueryDispatcher:
    def __init__(self, edges, verbose=False):
        """Construct inputs for APIQueryDispatcher
        
        params
        ======
        edges: an array of BTE edges with input added
        """
        self.edges = edges
        self.countAPIs()
        self.verbose = verbose
        if verbose:
            self.api_id = {}
            cnt = 1
            print(
                "BTE found {} APIs based on SmartAPI Meta-KG.\n".format(
                    len(self.api_cnt.keys())
                )
            )
            for k, v in self.api_cnt.items():
                print("API {}. {} ({} API calls)".format(cnt, k, v))
                self.api_id[k] = {"id": cnt, "current": 1}
                cnt += 1
            print("\n")

    def countAPIs(self):
        self.api_cnt = Counter()
        if self.edges:
            for edge in self.edges:
                self.api_cnt[edge["association"]["api_name"]] += 1

    @staticmethod
    def print_request(method, url, params, request_body, json_body=None):
        if params:
            url += "?"
            params = [(str(m) + "=" + str(n)) for m, n in params.items()]
            url += "&".join(params)
        if method == "get":
            return url
        if method == "post":
            if request_body:
                url += " (POST -d "
                url += str(request_body)
                url += ")"
            if json_body:
                url += " (POST -d "
                url += str(json_body)
                url += ")"
            return url
        return ""

    async def callSingleAPI(self, edge, session):
        qb = QueryBuilder(edge)
        if qb.config.get("json"):
            try:
                async with session.post(
                    qb.config.get("url"),
                    json=qb.config.get("json"),
                    timeout=ClientTimeout(8),
                ) as res:
                    if self.verbose:
                        print(
                            "API {}: {}".format(
                                str(self.api_id[edge["association"]["api_name"]]["id"])
                                + "."
                                + str(
                                    self.api_id[edge["association"]["api_name"]][
                                        "current"
                                    ]
                                ),
                                self.print_request(
                                    "post",
                                    qb.config.get("url"),
                                    qb.config.get("params"),
                                    qb.config.get("data"),
                                    qb.config.get("json"),
                                ),
                            )
                        )
                    if res.status >= 400:
                        if self.verbose:
                            print(
                                "API call to {} with input {} failed with status code {}".format(
                                    edge["association"]["api_name"],
                                    edge["input"],
                                    res.status,
                                )
                            )
                        return
                    try:
                        res = await res.json()
                    except JSONDecodeError:
                        res = await res.text()
                        res = json.loads(res)
                    tf = Transformer({"response": res, "edge": edge})
                    result = tf.transform()
                    if self.verbose:
                        print(
                            "API {} {}: {} hits".format(
                                str(self.api_id[edge["association"]["api_name"]]["id"])
                                + "."
                                + str(
                                    self.api_id[edge["association"]["api_name"]][
                                        "current"
                                    ]
                                ),
                                edge["association"]["api_name"],
                                len(result),
                            )
                        )
                    if "filter" in edge:
                        result = filter_response(result, edge["filter"])
                        if self.verbose:
                            print(
                                "API {} {}: {} hits after applying filters".format(
                                    str(
                                        self.api_id[edge["association"]["api_name"]][
                                            "id"
                                        ]
                                    )
                                    + "."
                                    + str(
                                        self.api_id[edge["association"]["api_name"]][
                                            "current"
                                        ]
                                    ),
                                    edge["association"]["api_name"],
                                    len(result),
                                )
                            )
                    if self.verbose:
                        self.api_id[edge["association"]["api_name"]]["current"] += 1
                    return result
            except asyncio.TimeoutError:
                if self.verbose:
                    print(
                        "API call to {} with input {} failed with timeout error. Current timeout limit is 5 seconds".format(
                            edge["association"]["api_name"], edge["input"]
                        )
                    )
                return
            except Exception as e:
                if self.verbose:
                    traceback.print_exc()
                    print(
                        "API call to {} with input {} failed with unknown response".format(
                            edge["association"]["api_name"], edge["input"]
                        )
                    )
                return
        elif qb.config.get("method") == "get":
            try:
                async with session.get(
                    qb.config.get("url"),
                    params=qb.config.get("params"),
                    timeout=ClientTimeout(8),
                ) as res:
                    if self.verbose:
                        print(
                            "API {}: {}".format(
                                str(self.api_id[edge["association"]["api_name"]]["id"])
                                + "."
                                + str(
                                    self.api_id[edge["association"]["api_name"]][
                                        "current"
                                    ]
                                ),
                                self.print_request(
                                    "get",
                                    qb.config.get("url"),
                                    qb.config.get("params"),
                                    qb.config.get("data"),
                                ),
                            )
                        )
                    if res.status >= 400:
                        if self.verbose:
                            print(
                                "API call to {} with input {} failed with status code {}".format(
                                    edge["association"]["api_name"],
                                    edge["input"],
                                    res.status,
                                )
                            )
                        return
                    try:
                        res = await res.json()
                    except JSONDecodeError:
                        res = await res.text()
                        res = json.loads(res)
                    tf = Transformer({"response": res, "edge": edge})
                    result = tf.transform()
                    if self.verbose:
                        print(
                            "API {} {}: {} hits".format(
                                str(self.api_id[edge["association"]["api_name"]]["id"])
                                + "."
                                + str(
                                    self.api_id[edge["association"]["api_name"]][
                                        "current"
                                    ]
                                ),
                                edge["association"]["api_name"],
                                len(result),
                            )
                        )
                    if "filter" in edge:
                        result = filter_response(result, edge["filter"])
                        if self.verbose:
                            print(
                                "API {} {}: {} hits after applying filters".format(
                                    str(
                                        self.api_id[edge["association"]["api_name"]][
                                            "id"
                                        ]
                                    )
                                    + "."
                                    + str(
                                        self.api_id[edge["association"]["api_name"]][
                                            "current"
                                        ]
                                    ),
                                    edge["association"]["api_name"],
                                    len(result),
                                )
                            )
                    if self.verbose:
                        self.api_id[edge["association"]["api_name"]]["current"] += 1
                    return result
            except asyncio.TimeoutError:
                if self.verbose:
                    print(
                        "API call to {} with input {} failed with timeout error. Current timeout limit is 5 seconds".format(
                            edge["association"]["api_name"], edge["input"]
                        )
                    )
                return
            except Exception as e:
                if self.verbose:
                    traceback.print_exc()
                    print(
                        "API call to {} with input {} failed with unknown response".format(
                            edge["association"]["api_name"], edge["input"]
                        )
                    )
                return
        elif qb.config.get("method") == "post":
            try:
                async with session.post(
                    qb.config.get("url"),
                    params=qb.config.get("params"),
                    data=qb.config.get("data"),
                    headers=qb.POST_HEADER,
                    timeout=ClientTimeout(20),
                ) as res:
                    if self.verbose:
                        print(
                            "API {}: {}".format(
                                str(self.api_id[edge["association"]["api_name"]]["id"])
                                + "."
                                + str(
                                    self.api_id[edge["association"]["api_name"]][
                                        "current"
                                    ]
                                ),
                                self.print_request(
                                    "post",
                                    qb.config.get("url"),
                                    qb.config.get("params"),
                                    qb.config.get("data"),
                                ),
                            )
                        )
                    if res.status >= 400:
                        if self.verbose:
                            print(
                                "API call to {} with input {} failed with status code {}".format(
                                    edge["association"]["api_name"],
                                    edge["input"],
                                    res.status,
                                )
                            )
                            print("config", qb.config)
                        return
                    res = await res.json()
                    tf = Transformer({"response": res, "edge": edge})
                    result = tf.transform()
                    if self.verbose:
                        print(
                            "API {} {}: {} hits".format(
                                str(self.api_id[edge["association"]["api_name"]]["id"])
                                + "."
                                + str(
                                    self.api_id[edge["association"]["api_name"]][
                                        "current"
                                    ]
                                ),
                                edge["association"]["api_name"],
                                len(result),
                            )
                        )
                    if "filter" in edge:
                        result = filter_response(result, edge["filter"])
                        if self.verbose:
                            print(
                                "API {} {}: {} hits after applying filters".format(
                                    str(
                                        self.api_id[edge["association"]["api_name"]][
                                            "id"
                                        ]
                                    )
                                    + "."
                                    + str(
                                        self.api_id[edge["association"]["api_name"]][
                                            "current"
                                        ]
                                    ),
                                    edge["association"]["api_name"],
                                    len(result),
                                )
                            )
                    if self.verbose:
                        self.api_id[edge["association"]["api_name"]]["current"] += 1
                    return result
            except asyncio.TimeoutError:
                if self.verbose:
                    print(
                        "API call to {} with input {} failed with timeout error. Current timeout limit is 20 seconds".format(
                            edge["association"]["api_name"], edge["input"]
                        )
                    )
                return
            except Exception as e:
                if self.verbose:
                    traceback.print_exc()
                    print(
                        "API call to {} with input {} failed with unknown response".format(
                            edge["association"]["api_name"], edge["input"]
                        )
                    )
                return

    async def annotate(self, res, session):
        output_ids = defaultdict(set)
        for item in res:
            output_ids[item["$association"]["output_type"]].add(item["$output"])
        resolved_ids = await asyncQuery(output_ids, session)
        for item in res:
            if item["$output"] in resolved_ids:
                item["$output_id_mapping"] = {
                    "resolved_ids": resolved_ids[item["$output"]],
                    "original": item["$output"],
                }
                item["label"] = resolved_ids[item["$output"]]["id"]["label"]
                item["id"] = item["$output"] = resolved_ids[item["$output"]]["id"][
                    "identifier"
                ]
            else:
                item["label"] = item["id"] = item["$output"]
        return res

    def scheduleQueryTasks(self, session):
        tasks = defaultdict(list)
        cnt = Counter()
        for edge in self.edges:
            if edge["query_operation"]["server"].startswith(
                "https://biothings.ncats.io"
            ):
                api = "pending API"
            else:
                api = edge["association"]["api_name"]
            cnt[api] += 1
            if api == "MyVariant.info API":
                tasks[cnt[api]].append(self.callSingleAPI(edge, session))
            if api == "pending API":
                task_id = floor(cnt[api] / 2)
                tasks[task_id].append(self.callSingleAPI(edge, session))
            else:
                task_id = floor(cnt[api] / MAX_CONCURRENT_QUERIES_ON_SINGLE_API)
                tasks[task_id].append(self.callSingleAPI(edge, session))
        return tasks

    async def asyncQuery(self):
        """Asynchronously make a list of API calls."""
        if self.verbose:
            print("==== Step #2: Query path execution ====\n")
            print(
                "NOTE: API requests are dispatched in parallel, so the list of APIs below is ordered by query time.\n"
            )
        responses = []
        timeout = ClientTimeout(total=5)
        async with ClientSession(
            timeout=timeout, connector=TCPConnector(ssl=False)
        ) as session:
            tasks = self.scheduleQueryTasks(session)
            for task in tasks.values():
                res = await asyncio.gather(*task)
                for _res in res:
                    if _res:
                        responses += _res
            if responses:
                if self.verbose:
                    print("\nBTE starts to perform id-to-object translation.\n")
                responses = await self.annotate(responses, session)
            return responses

    def syncQuery(self, loop=None):
        if not loop:
            loop = asyncio.get_event_loop()
        # return loop.create_task(self.asyncQuery())
        return loop.run_until_complete(self.asyncQuery())



from .biothings_transformer import BioThingsTransformer


class SemmedTransformer(BioThingsTransformer):
    def wrap(self, res):
        result = {}
        for pred, val in res.items():
            tmp = []
            if isinstance(val, list) and len(val) > 0:
                for item in val:
                    if item["@type"] == self.edge["association"]["output_type"] or (
                        item["@type"] == "DiseaseOrPhenotypicFeature"
                        and self.edge["association"]["output_type"] == "Disease"
                    ):
                        item["UMLS"] = item.pop("umls")
                        item["pubmed"] = item.pop("pmid")
                        tmp.append(item)
            if len(tmp) > 0:
                result[pred] = tmp
        return result

    def jsonTransform(self, res):
        return res



def restructure_ctd_response(json_doc):
    """restructure the JSON output from CTD API

    parameters
        * json_doc: the API response from CTD API
    """
    # convert from list to dict
    if isinstance(json_doc, list):
        json_doc = {'data': json_doc}
    for _doc in json_doc['data']:
        if "PubMedIds" in _doc:
            _doc["PubMedIds"] = _doc["PubMedIds"].split("|")
        if "DiseaseID" in _doc:
            _doc["DiseaseID"] = _doc["DiseaseID"].split(':')[-1]
    return json_doc



from collections import defaultdict

from .config import FILTER_FIELDS


def filterOps(ops, criteria):
    """filter an array of objects based on the filter criteria"""
    all_values = defaultdict(set)
    filters = {}
    for op in ops:
        for field in FILTER_FIELDS:
            all_values[field].add(op["association"].get(field))
    for field in FILTER_FIELDS:
        if field not in criteria or criteria.get(field) is None:
            filters[field] = all_values[field]
        else:
            if not isinstance(criteria[field], list):
                criteria[field] = [criteria[field]]
            filters[field] = set(criteria[field])
    res = []
    for op in ops:
        match = True
        for field in FILTER_FIELDS:
            if op["association"][field] not in filters[field]:
                match = False
                break
        if match:
            res.append(op)
    return res



SEMANTIC_TYPE_ID_MAPPING = {
    'ChemicalSubstance': ['CHEBI'],
    'Cell': ['CL'],
    'DiseaseOrPhenotypicFeature': ['DOID'],
    'Gene': ['HGNC'],
    'MolecularActivity': ['MOP', 'GO'],
    'BiologicalProcess': ['GO'],
    "Protein": ['PR'],
    "GenomicEntity": ['SO'],
    "AnatomicalEntity": ["UBERON"],
    "CellularComponent": ['GO']
}
cord = {
    "Gene": {
        "associated_with": [
            "ChemicalSubstance",
            "Cell",
            "DiseaseOrPhenotypicFeature",
            "MolecularActivity",
            "Protein",
            "GenomicEntity",
            "Gene",
            "AnatomicalEntity",
            "BiologicalProcess",
            "CellularComponent"
        ]
    },
    "Cell": {
        "associated_with": [
            "ChemicalSubstance",
            "Cell",
            "DiseaseOrPhenotypicFeature",
            "MolecularActivity",
            "Protein",
            "GenomicEntity",
            "Gene",
            "AnatomicalEntity",
            "BiologicalProcess",
            "CellularComponent"
        ]
    },
    "DiseaseOrPhenotypicFeature": {
        "associated_with": [
            "ChemicalSubstance",
            "Cell",
            "DiseaseOrPhenotypicFeature",
            "MolecularActivity",
            "Protein",
            "GenomicEntity",
            "Gene",
            "AnatomicalEntity",
            "BiologicalProcess",
            "CellularComponent"
        ]
    },
    "MolecularActivity": {
        "associated_with": [
            "ChemicalSubstance",
            "Cell",
            "DiseaseOrPhenotypicFeature",
            "MolecularActivity",
            "Protein",
            "GenomicEntity",
            "Gene",
            "AnatomicalEntity",
            "BiologicalProcess",
            "CellularComponent"
        ]
    },
    "Protein": {
        "associated_with": [
            "ChemicalSubstance",
            "Cell",
            "DiseaseOrPhenotypicFeature",
            "MolecularActivity",
            "Protein",
            "GenomicEntity",
            "Gene",
            "AnatomicalEntity",
            "BiologicalProcess",
            "CellularComponent"
        ]
    },
    "GenomicEntity": {
        "associated_with": [
            "ChemicalSubstance",
            "Cell",
            "DiseaseOrPhenotypicFeature",
            "MolecularActivity",
            "Protein",
            "GenomicEntity",
            "Gene",
            "AnatomicalEntity",
            "BiologicalProcess",
            "CellularComponent"
        ]
    },
    "AnatomicalEntity": {
        "associated_with": [
            "ChemicalSubstance",
            "Cell",
            "DiseaseOrPhenotypicFeature",
            "MolecularActivity",
            "Protein",
            "GenomicEntity",
            "Gene",
            "AnatomicalEntity",
            "BiologicalProcess",
            "CellularComponent"
        ]
    },
    "BiologicalProcess": {
        "associated_with": [
            "ChemicalSubstance",
            "Cell",
            "DiseaseOrPhenotypicFeature",
            "MolecularActivity",
            "Protein",
            "GenomicEntity",
            "Gene",
            "AnatomicalEntity",
            "BiologicalProcess",
            "CellularComponent"
        ]
    },
    "CellularComponent": {
        "associated_with": [
            "ChemicalSubstance",
            "Cell",
            "DiseaseOrPhenotypicFeature",
            "MolecularActivity",
            "Protein",
            "GenomicEntity",
            "Gene",
            "AnatomicalEntity",
            "BiologicalProcess",
            "CellularComponent"
        ]
    },
    "ChemicalSubstance": {
        "associated_with": [
            "ChemicalSubstance",
            "Cell",
            "DiseaseOrPhenotypicFeature",
            "MolecularActivity",
            "Protein",
            "GenomicEntity",
            "Gene",
            "AnatomicalEntity",
            "BiologicalProcess",
            "CellularComponent"
        ]
    }
}


from collections import defaultdict

from .base_transformer import BaseTransformer
from ....config_new import ALWAYS_PREFIXED


class ReasonerTransformer(BaseTransformer):
    def wrap(self, res):
        result = defaultdict(list)
        if not res.get("knowledge_graph"):
            return res
        if not res["knowledge_graph"].get("edges"):
            return res
        for edge in res["knowledge_graph"]["edges"]:
            if "target_id" in edge:
                prefix = edge["target_id"].split(":")[0]
                if prefix in ALWAYS_PREFIXED:
                    result[prefix].append(edge["target_id"])
                else:
                    result[prefix].append(edge["target_id"].split(":")[-1])
        return {self.edge["association"]["predicate"]: result}

    def jsonTransform(self, res):
        return res



from .base_transformer import BaseTransformer


class COHDTransformer(BaseTransformer):
    def wrap(self, res):
        print("cohd transformer triggerd")
        if not res.get("results"):
            return res
        new_res = {"results": []}
        for _doc in res["results"]:
            if _doc.get("adj_p-value") == 0:
                for k in _doc:
                    _doc[k] = str(_doc[k])
                new_res["results"].append(_doc)
        return new_res



from os.path import commonprefix


def generateCurie(idType, _id):
    if isinstance(_id, list):
        _id = _id[0]
    elif isinstance(_id, int):
        _id = str(_id)
    if ":" in _id:
        _id = _id.split(":")[-1]
    return idType + ":" + _id


def find_longest_common_path(paths) -> str:
    """Find longest common path among a list of paths.

    :param: paths: a list of paths, with '.' as the separator
    """
    return commonprefix(paths).rsplit(".", 1)[0]


def process_publications(res):
    if "pubmed" in res:
        if not isinstance(res["pubmed"], list):
            res["pubmed"] = [str(res["pubmed"])]
        res["publications"] = [
            ("PMID:" + str(item).strip("PMID:")) for item in res["pubmed"]
        ]
    if "pmc" in res:
        if not isinstance(res["pmc"], list):
            res["pmc"] = [str(res["pmc"])]
        res["publications"] = [
            ("PMC:" + str(item).strip("PMC:")) for item in res["pmc"]
        ]
    return res



# -*- coding: utf-8 -*-
"""A collection of util functions related to networkx

.. moduleauthor:: Jiwen Xin <kevinxin@scripps.edu>


"""

from collections import defaultdict


def load_res_to_networkx(_res, G, labels, id_mapping, output_id_types):
    """Load restructured API response into a networkx MultiDiGraph.

    Parameters
        * G: networkx MultiDiGraph
        * _res: restructured API response
        * labels: list of schema properties to extract from API response
        * id_mapping: dict containing mapping between equivalent ids and original ids
        * output_id_types: list of output identifiers
    """
    # check if API response is empty
    if not _res:
        return G
    # m represent input id, n represent parsed API output
    for input_id, parsed_api_output in _res.items():
        if not parsed_api_output:
            continue
        # a represent schema property, b represent value
        for prop, prop_vals in parsed_api_output.items():
            if prop not in labels:
                continue
            for _val in prop_vals:
                if not isinstance(_val, dict):
                    G.add_node(
                        str(_val),
                        identifier=prop,
                        type=parsed_api_output["@type"],
                        level=2,
                    )
                    G.add_edge(id_mapping[input_id], str(_val), info=None, label=prop)
                else:
                    for i, j in _val.items():
                        if i in output_id_types and j:
                            output_type = _val.get("@type")
                            source = _val.get("$source")
                            if not isinstance(j, list):
                                j = [j]
                            j = [str(jj) for jj in j]
                            G.add_nodes_from(j, identifier=i, type=output_type, level=2)
                            for _j in j:
                                G.add_edge(
                                    id_mapping[input_id],
                                    _j,
                                    info=_val,
                                    label=prop,
                                    source=source,
                                )
    return G


def add_equivalent_ids_to_nodes(G, IDResolver):
    """Add equivalent ids to each node.

    Parameters
        * G: Networkx Graph
        * IDConverter: Python Class in BTE to convert IDs

    TODO: This is weird, shouldn't include IDConverter in this.
    """
    # check if G is empty
    if not G:
        return (G, {})
    # get all nodes which are level 2 (output nodes)
    output_ids = [x for x, y in G.nodes(data=True) if y and y["level"] == 2]
    # check if there is no output nodes
    if not output_ids:
        return (G, {})
    # group output ids based on identifier and type
    idc_inputs = []
    output_ids_dict = defaultdict(list)
    for _id in output_ids:
        type_identifier = G.nodes[_id]["type"] + "," + G.nodes[_id]["identifier"]
        output_ids_dict[type_identifier].append(_id)
    # construct inputs for IDConverter
    for k, v in output_ids_dict.items():
        input_cls, input_id = k.split(",")
        idc_inputs.append((v, input_id, input_cls))
    # find equivalent ids
    equivalent_ids = IDResolver.resolve_ids(idc_inputs)
    # populate nodes with equivalent ids
    for m, n in equivalent_ids.items():
        G.nodes[m.split(":", 1)[-1]]["equivalent_ids"] = n
    return (G, equivalent_ids)


def merge_two_networkx_graphs(G1, G2):
    """Merge two networkx MultiDiGraphs.

    :param: G1: networkx graph as the source graph
    :param: G2: networkx graph added to G1

    TODO: line G1.add_edges_from(G2.edges(data=True)) will introduce duplicate edges
    """
    nodes_to_add = []
    for k, v in G2.nodes(data=True):
        if k not in G1:
            nodes_to_add.append((k, v))
    G1.add_nodes_from(nodes_to_add)
    G1.add_edges_from(G2.edges(data=True))
    return G1



import json
import yaml
import requests


def load_json_or_yaml(file_path):
    """Load either json or yaml document from file path or url or JSON doc.

    :param: file_path: The path of the url doc, could be url or file path
    """
    # handle json doc
    if isinstance(file_path, dict):
        return file_path
    # handle url
    if isinstance(file_path, str) and file_path.startswith("http"):
        with requests.get(file_path) as url:
            # check if http requests returns a success status code
            if url.status_code != 200:
                raise ValueError("Invalid URL!")
            _data = url.content
    # handle file path
    else:
        try:
            with open(str(file_path)) as f:
                _data = f.read()
        except FileNotFoundError:
            raise ValueError("Invalid File Path!")
    try:
        if isinstance(_data, bytes):
            _data = _data.decode('utf-8')
        data = json.loads(_data)
    except json.JSONDecodeError:   # for py>=3.5
        try:
            data = yaml.load(_data, Loader=yaml.SafeLoader)
        except (yaml.scanner.ScannerError,
                yaml.parser.ParserError):
            raise ValueError("Not a valid JSON or YAML format.")
    return data


from .biothings_transformer import BioThingsTransformer


class CordTransformer(BioThingsTransformer):
    def wrap(self, res):
        PREFIXES = ["pr", "go", "mop", "hgnc", "uberon", "so", "cl", "doid", "chebi"]
        result = {}
        for val in res.values():
            tmp = []
            if isinstance(val, list) and len(val) > 0:
                for item in val:
                    if item["@type"] == self.edge["association"]["output_type"] or (
                        item["@type"] == "DiseaseOrPhenotypicFeature"
                        and self.edge["association"]["output_type"] == "Disease"
                    ):
                        for k in item.keys():
                            if k in PREFIXES:
                                item[k.upper()] = item.pop(k)
                        tmp.append(item)
            if len(tmp) > 0:
                result["related_to"] = tmp
        return result

    def jsonTransform(self, res):
        return res



API_LIST = [
    "mygene",
    "ctd",
    "cord_gene",
    "cord_protein",
    "cord_chemical",
    "cord_disease",
    "cord_cell",
    "cord_molecular_activity",
    "cord_biological_process",
    "cord_cellular_component",
    "cord_anatomy",
    "cord_genomic_entity",
    "semmed_gene",
    "semmed_chemical",
    "semmed_disease",
    "semmed_biological_process",
    "semmed_anatomy",
    "semmed_phenotype",
    "scibite",
    "biolink",
    "opentarget",
    "dgidb",
    "mydisease",
    "mychem",
    "myvariant",
    "DISEASES",
    "scigraph",
    "pharos",
    "hmdb",
    "hetio",
    "chembio",
    "mgi_gene2phenotype",
    "ebi_gene2phenotype",
    "quickgo",
    "litvar",
    "ontology_lookup_service",
]

ID_RESOLVING_APIS = {
    "Gene": {
        "id_ranks": [
            "NCBIGene",
            "ENSEMBL",
            "HGNC",
            "UMLS",
            "UNIPROTKB",
            "SYMBOL",
            "OMIM",
            "MGI",
        ],
        "semantic": "Gene",
        "api_name": "mygene.info",
        "url": "https://mygene.info/v3",
        "mapping": {
            "NCBIGene": ["entrezgene"],
            "name": ["name"],
            "SYMBOL": ["symbol"],
            "UMLS": ["umls.cui", "umls.protein_cui"],
            "HGNC": ["HGNC"],
            "UNIPROTKB": ["uniprot.Swiss-Prot"],
            "ENSEMBL": ["ensembl.gene"],
            "OMIM": ["OMIM"],
            "MGI": ["MGI"],
        },
    },
    "SequenceVariant": {
        "id_ranks": ["DBSNP", "MYVARIANT_HG19", "HGVS", "ClinVar"],
        "api_name": "myvariant.info",
        "semantic": "SequenceVariant",
        "url": "https://myvariant.info/v1",
        "mapping": {
            "DBSNP": ["dbsnp.rsid", "clinvar.rsid", "dbnsfp.rsid"],
            "HGVS": [
                "clinvar.hgvs.genomic",
                "clinvar.hgvs.protein",
                "clinvar.hgvs.coding",
            ],
            "ClinVar": ["clinvar.rcv.accession"],
            "name": ["dbsnp.rsid", "clinvar.rsid", "dbnsfp.rsid"],
        },
    },
    "ChemicalSubstance": {
        "id_ranks": [
            "CHEBI",
            "CHEMBL.COMPOUND",
            "DRUGBANK",
            "PUBCHEM",
            "MESH",
            "UNII",
            "UMLS",
            "name",
            "CAS",
            "IUPAC",
            "formula",
        ],
        "semantic": "ChemicalSubstance",
        "api_name": "mychem.info",
        "url": "https://mychem.info/v1",
        "mapping": {
            "CHEMBL.COMPOUND": [
                "chembl.molecule_chembl_id",
                "drugbank.xrefs.chembl",
                "drugcentral.xrefs.chembl_id",
            ],
            "DRUGBANK": [
                "drugcentral.xrefs.drugbank_id",
                "pharmgkb.xrefs.drugbank",
                "chebi.xrefs.drugbank",
                "drugbank.id",
            ],
            "PUBCHEM": [
                "pubchem.cid",
                "drugbank.xrefs.pubchem.cid",
                "drugcentral.xrefs.pubchem_cid",
                "pharmgkb.xrefs.pubchem.cid",
            ],
            "CHEBI": [
                "chebi.id",
                "chembl.chebi_par_id",
                "drugbank.xrefs.chebi",
                "drugcentral.xrefs.chebi",
            ],
            "UMLS": ["drugcentral.xrefs.umlscui", "pharmgkb.xrefs.umls", "umls.cui"],
            "MESH": [
                "umls.mesh",
                "drugcentral.xrefs.mesh_descriptor_ui",
                "ginas.xrefs.MESH",
                "pharmgkb.xrefs.mesh",
            ],
            "UNII": [
                "drugcentral.xrefs.unii",
                "unii.unii",
                "aeolus.unii",
                "ginas.unii",
            ],
            "INCHIKEY": [
                "drugbank.inchi_key",
                "ginas.inchikey",
                "unii.inchikey",
                "chebi.inchikey",
            ],
            "INCHI": ["drugbank.inchi", "chebi.inchi", "chembl.inchi"],
            "KEGG": ["drugbank.xrefs.kegg.cid"],
            "name": [
                "chembl.pref_name",
                "drugbank.name",
                "umls.name",
                "ginas.preferred_name",
                "pharmgkb.name",
                "chebi.name",
                "drugbank.international_brands.name",
                "chembl.molecule_synonyms.synonyms",
                "drugbank.synonyms",
                "drugcentral.synonyms",
                "chebi.synonyms",
            ],
            "CAS": ["ginas.cas_primary", "pharmgkb.xrefs.cas", "chebi.xrefs.cas"],
            "IUPAC": ["pubchem.iupac.traditional", "drugbank.iupac"],
            "formula": [
                "chebi.formulae",
                "drugbank.formula",
                "pubchem.molecular_formula",
            ],
        },
    },
    "Disease": {
        "id_ranks": [
            "MONDO",
            "DOID",
            "OMIM",
            "ORPHANET",
            "UMLS",
            "MESH",
            "name",
            "OMOP",
        ],
        "semantic": "Disease",
        "api_name": "mydisease.info",
        "url": "http://mydisease.info/v1",
        "mapping": {
            "MONDO": ["mondo.mondo"],
            "DOID": ["mondo.xrefs.doid"],
            "UMLS": ["mondo.xrefs.umls", "disgenet.xrefs.umls"],
            "name": [
                "mondo.label",
                "disgenet.xrefs.disease_name",
                "mondo.synonym.exact",
                "mondo.synonym.related",
                "disease_ontology.synonyms.exact",
            ],
            "MESH": ["mondo.xrefs.mesh", "disease_ontology.xrefs.mesh", "ctd.mesh"],
            "OMIM": ["mondo.xrefs.omim", "hpo.omim"],
            "ORPHANET": ["hpo.orphanet", "mondo.xrefs.orphanet"],
            "OMOP": ["mondo.xrefs.cohd"],
        },
    },
    "PhenotypicFeature": {
        "id_ranks": [
            "UMLS",
            "SNOMEDCT",
            "HP",
            "MEDDRA",
            "EFO",
            "NCIT",
            "MESH",
            "MP",
            "name",
        ],
        "semantic": "PhenotypicFeature",
        "api_name": "HPO API",
        "url": "https://biothings.ncats.io/hpo",
        "mapping": {
            "UMLS": ["xrefs.umls"],
            "SNOMEDCT": ["xrefs.snomed_ct"],
            "HP": ["hp"],
            "MEDDRA": ["xrefs.meddra"],
            "EFO": ["xrefs.efo"],
            "NCIT": ["xrefs.ncit"],
            "MESH": ["xrefs.mesh"],
            "MP": ["xrefs.mp"],
            "name": ["name", "synonym.exact", "synonym.related", "synonym.broad"],
        },
    },
    "MolecularActivity": {
        "id_ranks": ["GO", "MetaCyc", "RHEA", "KEGG", "REACT", "name"],
        "semantic": "MolecularActivity",
        "api_name": "Gene Ontology Molecular Function API",
        "url": "https://biothings.ncats.io/go_mf",
        "mapping": {
            "GO": ["go"],
            "MetaCyc": ["xrefs.metacyc"],
            "RHEA": ["xrefs.rhea"],
            "KEGG": ["xrefs.kegg_reaction"],
            "REACT": ["xrefs.reactome"],
            "name": ["name", "synonym.exact", "synonym.related", "synonym.broad"],
        },
    },
    "BiologicalProcess": {
        "id_ranks": ["GO", "MetaCyc", "REACT", "KEGG", "name"],
        "semantic": "BiologicalProcess",
        "api_name": "Gene Ontology Biological Process API",
        "url": "https://biothings.ncats.io/go_bp",
        "mapping": {
            "GO": ["go"],
            "MetaCyc": ["xrefs.metacyc"],
            "KEGG": ["xrefs.kegg_pathway"],
            "REACT": ["xrefs.reactome"],
            "name": ["name", "synonym.exact", "synonym.related", "synonym.broad"],
        },
    },
    "CellularComponent": {
        "id_ranks": ["GO", "MetaCyc", "RHEA", "name"],
        "semantic": "CellularComponent",
        "api_name": "Gene Ontology Cellular Component API",
        "url": "https://biothings.ncats.io/go_cc",
        "mapping": {
            "GO": ["go"],
            "MetaCyc": ["xrefs.metacyc"],
            "RHEA": ["xrefs.rhea"],
            "name": ["name", "synonym.exact", "synonym.related", "synonym.broad"],
        },
    },
    "Pathway": {
        "id_ranks": ["REACT", "KEGG", "PHARMGKB", "WIKIPATHWAYS", "name"],
        "semantic": "Pathway",
        "api_name": "geneset API",
        "url": "https://biothings.ncats.io/geneset",
        "mapping": {
            "REACT": ["reactome"],
            "WIKIPATHWAYS": ["wikipathways"],
            "KEGG": ["kegg"],
            "PHARMGKB": ["pharmgkb"],
            "name": ["name"],
        },
    },
    "AnatomicalEntity": {
        "id_ranks": ["UBERON", "UMLS", "MESH", "NCIT", "name"],
        "semantic": "AnatomicalEntity",
        "api_name": "UBERON API",
        "url": "https://biothings.ncats.io/uberon",
        "mapping": {
            "UBERON": ["uberon"],
            "UMLS": ["xrefs.umls"],
            "MESH": ["xrefs.mesh"],
            "NCIT": ["xrefs.ncit"],
            "name": ["name", "synonym.exact", "synonym.related", "synonym.broad"],
        },
    },
    "Cell": {
        "id_ranks": ["CL", "NCIT", "MESH", "EFO", "name"],
        "semantic": "Cell",
        "api_name": "Cell Ontology API",
        "url": "https://biothings.ncats.io/cell_ontology",
        "mapping": {
            "CL": ["cl"],
            "NCIT": ["xrefs.ncit"],
            "MESH": ["xrefs.mesh"],
            "EFO": ["xrefs.efo"],
            "name": ["name", "synonym.exact", "synonym.related", "synonym.broad"],
        },
    },
}

MAX_CONCURRENT_QUERIES_ON_SINGLE_API = 5

ALWAYS_PREFIXED = [
    "RHEA",
    "GO",
    "CHEBI",
    "HP",
    "MONDO",
    "DOID",
    "EFO",
    "UBERON",
    "MP",
    "CL",
    "MGI",
]

SEMANTIC_TYPES_TO_EXPAND = [
    "Disease",
    "PhenotypicFeature",
    "AnatomicalEntity",
    "BiologicalProcess",
    "CellularComponent",
    "MolecularActivity",
]

EXPAND_API_LIST = [
    "mydisease.info API",
    "Gene Ontology Cellular Component API",
    "Gene Ontology Biological Process API",
    "UBERON Ontology API",
    "Gene Ontology Molecular Activity API",
    "Human Phenotype Ontology API",
]

BTE_FILTERS = ["nodeDegree", "ngd", "drugPhase", "survivalProbability"]



from .biothings_transformer import BioThingsTransformer


class MutFreqTransformer(BioThingsTransformer):
    def wrap(self, res):
        if "hits" not in res or not isinstance(res.get("hits"), list):
            return res
        for doc in res.get("hits"):
            if not doc.get("association"):
                continue
            if not doc["association"].get("freq_by_sample"):
                continue
            doc["association"]["freq_by_sample"] = str(
                doc["association"]["freq_by_sample"]
            )
        return res



from collections import defaultdict, Counter
from ...call_apis import APIQueryDispatcher
from ...smartapi_kg import MetaKG
from ..utils import annotateEdgesWithInput
from time import time


class SpecificityFilter:
    def __init__(self, stepResult, criteria, source_types):
        self.stepResult = stepResult
        self.criteria = criteria
        self.source_types = source_types
        self.kg = MetaKG()
        self.kg.constructMetaKG(source="local")

    def _collect_all_outputs(self):
        self.output_dict = {}
        for rec in self.stepResult:
            if rec["$association"]["output_type"] not in self.output_dict:
                self.output_dict[rec["$association"]["output_type"]] = {}
            self.output_dict[rec["$association"]["output_type"]][rec["$output"]] = rec[
                "$output_id_mapping"
            ]["resolved_ids"]

    def _find_smartapi_edges(self, output_type):
        return self.kg.filter(
            {"input_type": output_type, "output_type": self.source_types}
        )

    def _parse_response(self, response):
        cnt = {}
        summary = defaultdict(set)
        if not response:
            return {}
        for rec in response:
            summary[rec["$original_input"][rec["$input"]]["id"]["identifier"]].add(
                rec["$output"]
            )
        for k, v in summary.items():
            cnt[k] = len(v)
        return cnt

    def annotate(self):
        """
        Add specificity info to each edge.
        :param stepResult: a list of returned result from query.
        """
        print("start to annotate edges out")
        print("output types", self.source_types)
        t1 = time()
        if not self.stepResult:
            return
        self._collect_all_outputs()
        if not self.output_dict:
            return
        bte_edges = []
        for semanticType, resolvedIDs in self.output_dict.items():
            smartapi_edges = self._find_smartapi_edges(semanticType)
            if not smartapi_edges:
                continue
            tmp_edges = annotateEdgesWithInput(smartapi_edges, resolvedIDs.values())
            if not tmp_edges:
                continue
            bte_edges += tmp_edges
        print("found {} bte edges".format(len(bte_edges)))
        if not bte_edges:
            return

        dp = APIQueryDispatcher(bte_edges)
        res = dp.syncQuery()
        cnt = self._parse_response(res)
        for rec in self.stepResult:
            rec["$edgesOut"] = (
                cnt[rec.get("$output")] if rec.get("$output") in cnt else None
            )
        t2 = time()
        print("Annotating results took {} seconds".format(t2 - t1))
        return

    def filter(self):
        res = []
        if "sort" in self.criteria:
            limit = (
                int(self.criteria["sort"].get("limit"))
                if self.criteria["sort"].get("limit")
                else 0
            )
            if self.criteria["sort"].get("direction") != "descend":
                idList = [item[0] for item in self.cnt.most_common(limit)]
                for rec in self.stepResult:
                    if rec["$output"] in idList:
                        res.append(rec)
                return res
            sortedList = self.cnt.most_common(len(self.cnt))
            sortedList.reverse()
            idList = [item[0] for i, item in enumerate(sortedList) if i < limit]
            for rec in self.stepResult:
                if rec["$output"] in idList:
                    res.append(rec)
            return res



# -*- coding: utf-8 -*-
"""
Export the output of BTE to ReasonerAPIStd.

.. moduleauthor:: Jiwen Xin <kevinxin@scripps.edu>


"""
import hashlib
from collections import defaultdict


class ReasonerConverter:

    """Convert the output of BTE to ReasonerAPIStd."""

    def __init__(self):
        self.result = defaultdict(list)

    def load_bte_query_path(self, start, intermediate, end):
        """Load bte input query in the form of path.
        
        :param: start : the input of user query
        :param: intermediate : the intermediate nodes connecting input and output
        :param: end : the output of user query
        """
        self.path = [start.get("type")]
        self.start = start
        if ":" in self.start["primary"]["value"]:
            self.start_node_curie = self.start["primary"]["value"]
        else:
            self.start_node_curie = (
                self.start["primary"]["identifier"]
                + ":"
                + self.start["primary"]["value"]
            )

        if intermediate:
            if isinstance(intermediate, list):
                self.path += intermediate
            else:
                self.path.append(intermediate)
        if isinstance(end, str):
            self.path.append(end)
        elif isinstance(end, list):
            self.path.append(tuple(end))
        elif isinstance(end, dict):
            self.path.append(end.get("type"))

    def load_bte_output(self, G):
        """Load bte output in the format of networkx graph into class.
        
        :param: G : the networkx representation of bte output
        """
        self.G = G
        self.nodes = self.G.nodes(data=True)

    def get_curie(self, node):
        """Retrieve the curie representation of node.

        :param: node : the node id in networkx graph
        """
        if not node:
            return str(node)
        node_info = self.nodes[node]
        if "identifier" in node_info:
            prefix = node_info["identifier"]
            # if the node id is already in curie format
            # then no need to add prefix again
            if ":" in node:
                return node
            curie = prefix + ":" + node
            return curie
        return node

    @staticmethod
    def hash_id(_id):
        """Hash an id.

        :param: _id: a node or edge id
        """
        hash_obj = hashlib.sha256(str(_id).encode())
        return hash_obj.hexdigest()

    def fetch_edges(self):
        """Reorganize the edges into reasonerSTd format."""
        edges = []
        for k, v, o in self.G.edges(data=True):
            source_id = self.get_curie(k)
            target_id = self.get_curie(v)
            edge_source = o["info"].get("$api")
            _type = o.get("label")
            _id = self.hash_id(source_id + target_id + edge_source + _type)
            edge = {
                "source_id": source_id,
                "target_id": target_id,
                "edge_source": edge_source,
                "id": _id,
                "type": _type,
            }
            self.result[source_id + "|||" + target_id].append(_id)
            edges.append(edge)
        return edges

    def fetch_nodes(self):
        """Reorganize the nodes into reasonerSTD format."""
        nodes = []
        for k, v in self.nodes:
            name = v["equivalent_ids"].get("name")
            if name and isinstance(name, list):
                name = str(name[0])
            else:
                name = str(self.get_curie(k))
            node = {
                "id": self.get_curie(k),
                "name": name,
                "type": v["type"],
                "equivalent_identifiers": v["equivalent_ids"],
            }
            nodes.append(node)
        return nodes

    def generate_knowledge_graph(self):
        """Reorganize the nodes and edges into reasonerSTD format."""
        if len(self.G) == 0:
            return {"nodes": [], "edges": []}
        return {"nodes": self.fetch_nodes(), "edges": self.fetch_edges()}

    def generate_question_graph(self):
        if not self.path:
            return {"edges": [], "nodes": []}
        node_id = 0
        edge_id = 0

        nodes = []
        edges = []

        node2idmapping = {}

        for i, node in enumerate(self.path):
            if isinstance(node, str):
                nodes.append({"id": "n" + str(node_id), "type": node})
                node2idmapping[str(i) + "-" + node + "-0"] = "n" + str(node_id)
                node_id += 1
            elif isinstance(node, tuple):
                for j, _n in enumerate(node):
                    nodes.append({"id": "n" + str(node_id), "type": _n})
                    node2idmapping[str(i) + "-" + _n + "-" + str(j)] = "n" + str(
                        node_id
                    )
                    node_id += 1
            nodes[0]["curie"] = self.start_node_curie

        for i in range(0, len(self.path) - 1):
            source_node = self.path[i]
            target_node = self.path[i + 1]
            if isinstance(source_node, str):
                source_node = [source_node]
            if isinstance(target_node, str):
                target_node = [target_node]
            for p, _s in enumerate(source_node):
                for q, _t in enumerate(target_node):
                    source_id = node2idmapping[str(i) + "-" + _s + "-" + str(p)]
                    target_id = node2idmapping[str(i + 1) + "-" + _t + "-" + str(q)]
                    edges.append(
                        {
                            "id": "e" + str(edge_id),
                            "source_id": source_id,
                            "target_id": target_id,
                            "directed": True,
                        }
                    )
                    edge_id += 1
        return {"edges": edges, "nodes": nodes}

    def generate_result(self):
        result = {"node_bindings": [], "edge_bindings": []}
        if not self.result:
            return result
        for k, v in self.result.items():
            source_id, target_id = k.split("|||")
            if source_id == "name:{}".format(self.start["name"]):
                result["node_bindings"].append({"n0": source_id, "n1": target_id})
                result["edge_bindings"].append({"e0": v})
            else:
                result["node_bindings"].append({"n1": source_id, "n2": target_id})
                result["edge_bindings"].append({"e1": v})
        return result

    def generate_reasoner_response(self):
        """Generate reasoner response."""
        response = {
            "query_graph": self.generate_question_graph(),
            "knowledge_graph": self.generate_knowledge_graph(),
        }
        response["results"] = self.generate_result()
        return response



from .nodeDegree import NodeDegreeFilter
from .ngd import NGDFilter
from .drugPhase import DrugPhaseFilter
from .hypothesis import HypothesisFilter
from .specificity import SpecificityFilter
from functools import partial

BTE_FILTERS = ["nodeDegree", "ngd", "drugPhase", "survivalProbability", "edgesOut"]


class Filter:
    def __init__(self, stepResult, criteria, source_types):
        self.stepResult = stepResult
        self.criteria = criteria
        self.source_types = source_types
        self.annotated = False

    def annotate(self):
        f = NodeDegreeFilter(self.stepResult, {})
        f.annotateNodeDegree()
        if "ngd" in self.criteria:
            f = NGDFilter(self.stepResult, {})
            f.annotateNGD()
        if "edgesOut" in self.criteria:
            f = SpecificityFilter(self.stepResult, {}, self.source_types)
            f.annotate()
        if "drugPhase" in self.criteria:
            f = DrugPhaseFilter(self.stepResult, {})
            f.annotate()
        if "survivalProbability" in self.criteria:
            f = HypothesisFilter(self.stepResult, {})
            f.annotate()
        self.annotated = True
        return self.stepResult

    def filter_response(self):
        """
        Filter API response based on filtering criteria
        :param res: API Response
        :param criteria: filtering criteria
        """

        def filter_by_operation(rec, key, val, operation):
            if rec.get(key):
                if isinstance(rec.get(key), list):
                    rec[key] = rec[key][0]
                try:
                    if operation == "=" and type(val)(rec[key]) == val:
                        return True
                    if operation == ">" and type(val)(rec[key]) > val:
                        return True
                    if operation == "<" and type(val)(rec[key]) < val:
                        return True
                    return False
                except (ValueError, TypeError):
                    return False
            return False

        if not self.annotated:
            self.annotate()
        if (
            not self.stepResult
            or not isinstance(self.stepResult, list)
            or not len(self.stepResult) > 0
        ):
            return self.stepResult
        if not isinstance(self.criteria, dict):
            return self.stepResult
        for f, v in self.criteria.items():
            if not isinstance(v, dict):
                continue
            if f in BTE_FILTERS:
                f = "$" + f
                if "=" in v:
                    self.stepResult = list(
                        filter(
                            partial(
                                filter_by_operation, key=f, val=v["="], operation="="
                            ),
                            self.stepResult,
                        )
                    )
                    continue
                if ">" in v:
                    self.stepResult = list(
                        filter(
                            partial(
                                filter_by_operation, key=f, val=v[">"], operation=">"
                            ),
                            self.stepResult,
                        )
                    )
                elif "<" in v:
                    self.stepResult = list(
                        filter(
                            partial(
                                filter_by_operation, key=f, val=v["<"], operation="<"
                            ),
                            self.stepResult,
                        )
                    )
        return self.stepResult



from collections import defaultdict
from jsonpath_rw import parse
from .utils import find_longest_common_path


class Transformer:
    """Transform the JSON output from API response into schema format."""

    def __init__(self, json_doc, mapping):
        """
        Load json doc and transform mapping file.
        
        :param: json_doc: dict, the json document
        :param: mapping: dict, the schema mapping file which the transformation will be based on
        """
        self.json_doc = json_doc
        self.mapping = mapping

    @staticmethod
    def generate_parser(path):
        """
        Create path for jsonpath_rw.

        :param: path: jsonpath_rw path

        example: ensembl.gene -> ensembl[*].gene[*]
        """
        return parse((".").join([(i + "[*]") for i in path.split(".")]))

    @staticmethod
    def fetch_all_paths_from_mapping_file(mapping_file):
        """Fetch all paths from schema mapping file."""
        paths = []
        if not mapping_file or not isinstance(mapping_file, dict):
            return []
        for v in mapping_file.values():
            if isinstance(v, list):
                paths += v
            elif isinstance(v, str):
                paths.append(v)
            else:
                raise ValueError(
                    "The data type of each path should be either list or str"
                )
        return paths

    @staticmethod
    def group_jsonpaths(commonprefix, jsonpath_dict):
        result = defaultdict(list)
        # find the last common element, e.g. if 'go.BP' is common prefix, then 'BP' is the last common element
        lst_common_elem = commonprefix.rpartition(".")[2]
        for k, v in jsonpath_dict.items():
            for _path in v:
                splitted_path = _path.split(".")
                lst_common_element_idx = splitted_path.index(lst_common_elem)
                result[".".join(splitted_path[0 : lst_common_element_idx + 2])].append(
                    (k, _path)
                )
        return dict(result)

    @staticmethod
    def fetch_value_by_jsonpath(json_dict, json_path):
        path_elements = json_path.split(".")
        result = json_dict
        for _ele in path_elements:
            if not _ele.startswith("["):
                result = result[_ele]
            else:
                if isinstance(result, list):
                    result = result[_ele]
        return result

    @staticmethod
    def find_key_by_value(json_dict, value):
        """Retrieve the key of a python dict based on the value"""
        for k, v in json_dict.items():
            if isinstance(v, list):
                if value in v:
                    return k
            else:
                if v == value:
                    return k

    def parse_dict(self, mapping_dict):
        """
        mapping: {"dd": "a.b.c", "ee": "a.b.d"}
        case 1: {"a": {"b": {"c": 1, "d": 2}}}
        result 1: {"dd": 1, "ee": 2}
        case 2: {"a": [{"b": {"c": 1, "d": 2}}, {"b": {"c": 3, "d": 4}}]}
        result 2: [{"dd": 1, "ee": 2}]
        case 3: {"a": }
        """
        all_paths = self.fetch_all_paths_from_mapping_file(mapping_dict)
        common_path = find_longest_common_path(all_paths)
        # Handle cases where a common prefix could be found among all paths
        if common_path:
            paths_jsonpaths_dict = {}
            jsonpaths_values_dict = {}
            for path in all_paths:
                parser = self.generate_parser(path)
                parsing_result = parser.find(self.json_doc)
                jsonpaths = [str(match.full_path) for match in parsing_result]
                values = [match.value for match in parsing_result]
                jsonpaths_values_dict.update(dict(zip(jsonpaths, values)))
                paths_jsonpaths_dict.update({path: jsonpaths})
            grouped_jsonpaths = self.group_jsonpaths(common_path, paths_jsonpaths_dict)
            result = []
            for v in grouped_jsonpaths.values():
                _result = defaultdict(list)
                for _tuple in v:
                    schema_prop = self.find_key_by_value(mapping_dict, _tuple[0])
                    _result[schema_prop].append(jsonpaths_values_dict[_tuple[1]])
                result.append(dict(_result))
            return result
        # Handle cases where all paths doesn't share a common prefix
        result = {}
        for k, v in mapping_dict.items():
            if isinstance(v, str):
                parser = self.generate_parser(v)
                result[k] = self.fetch_value_from_single_path(parser)
            elif isinstance(v, list):
                _res = []
                for path in v:
                    parser = self.generate_parser(path)
                    _res += self.fetch_value_from_single_path(parser)
                result[k] = _res
        return result

    def fetch_value_from_single_path(self, parser):
        return [match.value for match in parser.find(self.json_doc)]

    def fetch_value(self, key, paths):
        """Fetch value from API response.
        
        :param: key: the biolink property name.
        :param: paths: the path to the field in API response.
        """
        # Handle only one field path correspond to the key.
        if isinstance(paths, str):
            parser = self.generate_parser(paths)
            return self.fetch_value_from_single_path(parser)
        # Handle more than one field paths correspond to the key.
        if isinstance(paths, list):
            result = []
            for path in paths:
                if isinstance(path, dict):
                    _res = self.parse_dict(path)
                elif isinstance(path, str):
                    parser = self.generate_parser(path)
                    _res = self.fetch_value_from_single_path(parser)
                else:
                    raise ValueError("{} is not valid".format(path))
                if isinstance(_res, list):
                    result += _res
                else:
                    result.append(_res)
            return result
        if isinstance(paths, dict):
            return self.parse_dict(paths)
        return None

    def transform(self):
        new_json_doc = {k: self.fetch_value(k, v) for k, v in self.mapping.items()}
        return new_json_doc



"""Parse the biothings schema"""

from .config import BIOTHINGS_SCHEMA_URL, PREFIX_TO_REMOVE
from .utils.dataload import load_json_or_yaml
from .utils.common import remove_prefix

class SchemaParser():
    def __init__(self):
        self.schema_json = remove_prefix(load_json_or_yaml(BIOTHINGS_SCHEMA_URL),
                                         PREFIX_TO_REMOVE)
        self.properties = {}
        self.ids = []
        self.clses = []
        self.process_schema()

    def process_schema(self):
        for rec in self.schema_json['@graph']:
            if "rdfs:subPropertyOf" in rec and rec["rdfs:subPropertyOf"]["@id"] == "http://schema.org/identifier":
                self.ids.append(rec["@id"])
            elif rec["@type"] == "rdf:Property":
                self.properties[rec["@id"]] = {"inverse_property": None}
                if "schema:inverseOf" in rec:
                    self.properties[rec["@id"]]["inverse_property"] = rec["schema:inverseOf"]["@id"]
            elif rec["@type"] == "rdfs:Class":
                self.clses.append(rec["@id"])



from collections import Counter
from data_tools import plotting as dtp


def display_graph(table):
    cnt = Counter()
    Row_list = []
    label_dict = {}
    df = table[
        [item for item in table.columns if "label" in item or "type" in item]
    ].drop_duplicates()
    for rows in df.itertuples():
        if len(Row_list) > 12:
            break
        if cnt[rows.output_label] < 3:
            my_list = []

            for i in range(1, len(list(rows)), 2):
                my_list.append(list(rows)[i])
                label_dict[list(rows)[i]] = list(rows)[i + 1]
            # Create list for the current row
            # my_list = list(rows)[1:]

            # append the list to the final list
            Row_list.append(my_list)

            cnt[rows.output_label] += 1
    G = dtp.build_explanitory_graph(Row_list, node_id_to_label=label_dict)
    dtp.draw_explanitory_graph(G)



def generateFailedResponse(curie, _type):
    return {
        "id": {"identifier": curie, "label": curie},
        "curies": [curie],
        "db_ids": {curie.split(":")[0]: [curie]},
        "type": _type,
        "flag": "failed",
    }



from copy import deepcopy

from ..config_new import ALWAYS_PREFIXED


class QueryBuilder:
    def __init__(self, edge):
        self.POST_HEADER = {"content-type": "application/x-www-form-urlencoded"}
        self.edge = edge
        self.server = edge["query_operation"]["server"]
        if edge["query_operation"]["server"].endswith("/"):
            self.server = self.server.strip("/")
        self.url = self.server + edge["query_operation"]["path"]
        self.method = edge["query_operation"]["method"]
        self.supportBatch = edge["query_operation"].get("supportBatch")
        self.input = edge.get("input")
        self.inputSeparator = edge["query_operation"].get("inputSeparator")
        self.params = deepcopy(edge["query_operation"].get("params"))
        self.data = ""
        if "reasoner" in edge["query_operation"].get("tags"):
            self.constructRequestConfigForReasonerAPI()
        else:
            self.constructInput()
            self.constructRequestBody()
            self.constructParams()
            self.constructRequestConfig()

    def constructInput(self):
        """Construct input based on method and inputSeparator"""
        if self.supportBatch:
            if not isinstance(self.input, list):
                self.input = [self.input]
            self.input = self.inputSeparator.join(self.input)

    def constructParams(self):
        """Construct parameters for API calls"""
        if self.edge["query_operation"].get("path_params"):
            for param in self.edge["query_operation"].get("path_params"):
                self.url = self.url.replace(
                    "{" + param + "}", self.params.pop(param)
                ).replace("{inputs[0]}", self.input)
        for param, val in self.params.items():
            if isinstance(val, bool):
                self.params[param] = str(val)
            if isinstance(val, str):
                self.params[param] = val.replace("{inputs[0]}", self.input)

    def constructRequestBody(self):
        """Construct request body for API calls"""
        if self.edge["query_operation"].get("request_body") and "body" in self.edge[
            "query_operation"
        ].get("request_body"):
            body = self.edge["query_operation"]["request_body"]["body"]
            self.data = ""
            for k, v in body.items():
                self.data = (
                    self.data + k + "=" + v.replace("{inputs[0]}", self.input) + "&"
                )
            self.data = self.data.strip("&")

    def constructRequestConfig(self):
        """Construct the request config for python reqeust library."""
        self.config = {
            "url": self.url,
            "params": self.params,
            "data": self.data,
            "method": self.method,
            "timeout": 5000,
        }

    def constructRequestConfigForReasonerAPI(self):
        """Construct the ReasonerStdAPI request config for python reqeust library."""

        def id2curie(prefix, val):
            if prefix in ALWAYS_PREFIXED:
                return val
            return prefix + ":" + val

        predicate = self.edge["association"]["predicate"]
        input_type = self.edge["association"]["input_type"].lower()
        output_type = self.edge["association"]["output_type"].lower()
        _input = id2curie(self.edge["association"]["input_id"], self.input)
        if self.edge["association"]["api_name"] == "Genetics Provider API":
            predicate = "associated"
            if input_type == "phenotypicfeature":
                input_type = "phenotype"
            if output_type == "phenotypicfeature":
                output_type = "phenotype"
        self.config = {
            "url": self.url,
            "json": {
                "message": {
                    "query_graph": {
                        "edges": [
                            {
                                "id": "e00",
                                "source_id": "n00",
                                "target_id": "n01",
                                "type": predicate,
                            }
                        ],
                        "nodes": [
                            {"curie": _input, "id": "n00", "type": input_type},
                            {"id": "n01", "type": output_type},
                        ],
                    }
                }
            },
        }



from .registry import Registry
from .config import metadata
from collections import defaultdict
import networkx as nx


class Metadata():
    """ Metadata Info for the Meta Knowledge Graph stored in BTE
    """
    def __init__(self, reg=None):
        if not reg:
            self.registry = Registry()
        else:
            self.registry = reg

    def list_all_semantic_types(self):
        """List all semantic types used in BTE"""
        semmantic_types = set()
        for _, _, info in self.registry.G.edges(data=True):
            semmantic_types.add(info['input_type'])
            semmantic_types.add(info['output_type'])
        return list(semmantic_types)

    def list_all_predicates(self):
        """List all predicates used in BTE"""
        predicates = set()
        for _, _, info in self.registry.G.edges(data=True):
            predicates.add(info['label'])
        return list(predicates)

    def list_all_id_types(self):
        """List all identifiers used in BTE"""
        return list(set(self.registry.G.nodes()))

    def list_all_associations(self):
        """List all associations used in BTE
        Each association is a triple with subject, predicate and object
        """
        associations = set()
        for p, o, info in self.registry.G.edges(data=True):
            _assoc = info['input_type'] + '|' + info['label'] + '|' + info['output_type']
            associations.add(_assoc)
        results = []
        for _assoc in associations:
            s, p, o = _assoc.split('|')
            results.append((s, p, o))
        return results

    def semantic_network_graph(self, edge="pred"):
        """Convert the meta knowledge graph into a semantic graph"""
        _id = 1
        id_dict = {}
        edges = set()
        edges_dict = defaultdict(list)
        result = {'nodes': [], 'edges': []}
        for _, _, info in self.registry.G.edges(data=True):
            input_type = info['input_type']
            output_type = info['output_type']
            predicate = info['label']
            if input_type not in id_dict:
                id_dict[input_type] = _id
                result['nodes'].append({'id': _id,
                                        'group': _id,
                                        'label': input_type})
                _id += 1
            if output_type not in id_dict:
                id_dict[output_type] = _id
                result['nodes'].append({'id': _id,
                                        'group': _id,
                                        'label': output_type})
                _id += 1
            if edge == 'pred':
                _edge = str(id_dict[input_type]) + predicate + str(id_dict[output_type])
                if _edge not in edges:
                    result['edges'].append({'from': id_dict[input_type],
                                        'to': id_dict[output_type],
                                        'label': predicate})
                    edges.add(_edge)
            else:
                edge_key = [id_dict[input_type], id_dict[output_type]]
                edge_key.sort()
                edge_key = '|'.join([str(i) for i in edge_key])
                edges_dict[edge_key].append(metadata[info['api']]['api_name'])
        if edge != 'pred':
            for k, v in edges_dict.items():
                _input_type, _output_type = k.split('|')
                v = list(set(v))
                result['edges'].append({'from': int(_input_type),
                                        'to': int(_output_type),
                                        'label': '\n'.join(v)})
        return result

    def semantic_network_nx(self, edge="pred"):
        """Convert the meta knowledge graph into a semantic networkx graph"""
        _id = 1
        G = nx.MultiDiGraph()
        edges = set()
        nodes = set()
        for _, _, info in self.registry.G.edges(data=True):
            input_type = info['input_type']
            if input_type not in nodes:
                nodes.add(input_type)
                G.add_node(input_type, label=input_type)
            output_type = info['output_type']
            if output_type not in nodes:
                nodes.add(output_type)
                G.add_node(output_type, label=output_type)
            api = info['api']
            if api.startswith("semmed"):
                api = 'semmed'
            if api.startswith("cord"):
                api = 'cord'
            edge = input_type + '-' + api + '-' + output_type
            if edge not in edges:
                edges.add(edge)
                edge_reverse = output_type + '-' + api + '-' + input_type
                edges.add(edge_reverse)
                G.add_edge(input_type, output_type, label=api, id='e' + str(_id))
                _id += 1
        return G

    def id_network_graph(self, edge="pred"):
        _id = 1
        id_dict = {}
        edges = set()
        result = {'nodes': [], 'edges': []}
        for _, _, info in self.registry.G.edges(data=True):
            input_type = info['input_id']
            output_type = info['output_id']
            predicate = info['label']
            if input_type not in id_dict:
                id_dict[input_type] = _id
                result['nodes'].append({'id': _id,
                                        'group': _id,
                                        'label': input_type})
                _id += 1
            if output_type not in id_dict:
                id_dict[output_type] = _id
                result['nodes'].append({'id': _id,
                                        'group': _id,
                                        'label': output_type})
                _id += 1
            if edge == 'pred':
                _edge = str(id_dict[input_type]) + predicate + str(id_dict[output_type])
                if _edge not in edges:
                    result['edges'].append({'from': id_dict[input_type],
                                        'to': id_dict[output_type],
                                        'label': predicate})
                    edges.add(_edge)
            else:
                _edge = str(id_dict[input_type]) + metadata[info['api']]['api_name'] + str(id_dict[output_type])
                if _edge not in edges:
                    result['edges'].append({'from': id_dict[input_type],
                                            'to': id_dict[output_type],
                                            'label': metadata[info['api']]['api_name']})
                    edges.add(_edge)
        return result



# -*- coding: utf-8 -*-
"""
Storing metadata information and connectivity of APIs.

.. moduleauthor:: Jiwen Xin <kevinxin@scripps.edu>


"""
import networkx as nx
from .config_new import API_LIST
from .smartapi_parser import SmartAPIParser
from pathlib import Path
CURRENT_PATH = Path(__file__)

class Registry():

    """Convert metadata information of APIs into a networkx graph."""

    def __init__(self):
        """Initialize networkx graph and load biothings apis."""
        self.G = nx.MultiDiGraph()
        self.smp = SmartAPIParser()
        self.registry = {}
        self.api_list = API_LIST
        self.load_smartapis()
        self.all_edges_info = self.G.edges(data=True)
        self.all_labels = {d[-1]['label'] for d in self.all_edges_info}
        self.all_inputs = {d[-1]['input_type'] for d in self.all_edges_info}
        self.all_outputs = {d[-1]['output_type'] for d in self.all_edges_info}

    def _reload_registry(self):
        """Reconstruct BTE Registry.
        
        note: This one is called everytime the API_LIST changes.
        """
        self.G = nx.MultiDiGraph()
        self.load_smartapis()
        self.all_edges_info = self.G.edges(data=True)
        self.all_labels = {d[-1]['label'] for d in self.all_edges_info}
        self.all_inputs = {d[-1]['input_type'] for d in self.all_edges_info}
        self.all_outputs = {d[-1]['output_type'] for d in self.all_edges_info}

    def show_all_apis(self):
        """List all APIs in the BTE Registry."""
        return self.api_list

    def remove_apis(self, apis):
        """Remove one or a list of APIs from registry
        
        :param: apis: list of APIs to be removed from BTE Registry.
        """
        if not isinstance(apis, list):
            apis = [apis]
        for api in apis:
            if api in self.api_list:
                self.api_list.remove(api)
                print("{} has been successfully removed!".format(api))
            else:
                raise ValueError("{} is not in the API list.".format(api))
        self._reload_registry()

    def refine_api_list(self, apis):
        """Set the API List.
        
        :params: apis: list of APIs to include in the BTE Registry
        """
        if not isinstance(apis, list):
            apis = [apis]
        self.api_list = []
        for api in apis:
            if api in API_LIST:
                self.api_list.append(api)
            else:
                self.api_list = API_LIST
                self._reload_registry()
                raise ValueError("{} is not in the API list.".format(api))
        self._reload_registry()

    def load_smartapis(self):
        """Load biothings API into registry network graph."""
        # load biothings schema
        # loop through API metadata
        for api in self.api_list:
            smartapi_file = Path.joinpath(CURRENT_PATH.parent,
                                          'smartapi/new_specs', api + '.json')
            self.smp.load_spec(smartapi_file)
            ops = self.smp.fetch_endpoint_info()
            for op in ops:
                self.G.add_edge(
                    op['input_id'],
                    op['output_id'],
                    operation_id=op['id'],
                    label=op['predicate'],
                    api=api,
                    api_type=op['api_type'],
                    input_type=op['input_type'],
                    output_type=op['output_type'],
                    operation=op
                )
        return self.G

    def filter_edges(self, input_cls=None, output_cls=None, edge_label=None):
        """
        Filter edges based on input, output and label.

        The relationship between bio-entities is represented as a networkx MultiDiGraph \
            in BioThings explorer. This function helps you filter for the relationships of your interest based on input/output/edge info.

        :param: input_cls (str|list|None) : the semantic type(s) of the input.
                   Optional
        :param: output_cls (str|list|None) : the semantic type(s) of the output.
                    Optional
        :param: edge_label (str|list|None) : the relationship between input and output.

        """
        if edge_label:
            if isinstance(edge_label, str):
                edge_label = [edge_label]
        # if no edge label is specified, set it as all labels
        else:
            edge_label = self.all_labels
        # if no input_cls is specified, set it as all input types
        if not input_cls:
            input_cls = self.all_inputs
        # if input_cls is str, convert it to list of one element
        elif isinstance(input_cls, str):
            input_cls = [input_cls]
        # if no output_cls is specified, set it as all output types
        if not output_cls:
            output_cls = self.all_outputs
        elif isinstance(output_cls, str):
            output_cls = [output_cls]
        return [d for u,v,d in self.all_edges_info if d['input_type'] in input_cls and d['output_type'] in output_cls and d['label'] in edge_label]



from aiohttp import ClientSession, TCPConnector
import asyncio

from .dispatch import Dispatcher
from .utils import generateFailedResponse


async def asyncQuery(inputIDs, session=None):
    """Asynchronously make a list of API calls."""
    responses = {}
    if not session:
        async with ClientSession(connector=TCPConnector(ssl=False)) as session:
            dp = Dispatcher(inputIDs, session)
            dp.dispatch()
            for task in dp.tasks.values():
                res = await asyncio.gather(*task)
                for _res in res:
                    responses.update(_res)
            return responses
    else:
        dp = Dispatcher(inputIDs, session)
        dp.dispatch()
        if dp.invalid and isinstance(dp.invalid, dict):
            for semantic_type, curies in dp.invalid.items():
                if curies and len(curies) > 0:
                    for curie in curies:
                        responses[curie] = generateFailedResponse(curie, semantic_type)
        for task in dp.tasks.values():
            res = await asyncio.gather(*task)
            for _res in res:
                if _res:
                    responses.update(_res)
        return responses


def syncQuery(inputIDs, loop=None):
    if not loop:
        loop = asyncio.new_event_loop()
    return loop.run_until_complete(asyncQuery(inputIDs))



from collections import Counter, defaultdict
import requests
import json


TYPE_TO_ID_MAPPING = {
    "Gene": "NCBIGene",
    "ChemicalSubstance": "CHEBI",
    "AnatomicalEntity": "UBERON",
    "BiologicalProcess": "GO",
    "MolecularActivity": "GO",
    "Cell": "CL",
    "SequenceVariant": "SO",
    "Disease": "MONDO",
    "PhenotypicFeature": "HP",
}


class NGDFilter:
    def __init__(self, stepResult, criteria):
        self.stepResult = stepResult
        self.cnt = Counter()
        self.criteria = criteria

    @staticmethod
    def _extractInputID(resolvedIDs, semanticType):
        if semanticType not in TYPE_TO_ID_MAPPING:
            return
        if "db_ids" not in resolvedIDs or not isinstance(resolvedIDs["db_ids"], dict):
            return
        prefix = TYPE_TO_ID_MAPPING[semanticType]
        if not prefix in resolvedIDs["db_ids"]:
            return
        return resolvedIDs["db_ids"][prefix]

    @staticmethod
    def _queryNGD(inputs):
        if not inputs:
            return
        result = []
        for i in range(0, len(inputs), 1000):
            a = {
                "q": [item.split("-") for item in inputs[i : i + 1000]],
                "scopes": [["subject.id", "object.id"], ["object.id", "subject.id"],],
                "fields": "association.ngd",
                "dotfield": True,
            }
            res = requests.post(
                "https://biothings.ncats.io/text_mining_co_occurrence_kp/query", json=a,
            )
            res = res.json()
            result += res
        return result

    @staticmethod
    def _parseResponse(res):
        if not res:
            return
        result = {}
        for rec in res:
            if "association.ngd" in rec:
                result["-".join(rec["query"])] = rec["association.ngd"]
        return result

    def annotateNGD(self):
        """
        Add node degree info to each edge.
        :param stepResult: a list of returned result from query.
        """
        id_dict = defaultdict(list)
        ngd_inputs = set()
        if not self.stepResult:
            return
        for i, rec in enumerate(self.stepResult):
            if not "$association" in rec:
                continue
            input_type = rec["$association"]["input_type"]
            input_resolved_ids = rec["$original_input"][rec["$input"]]
            output_type = rec["$association"]["output_type"]
            output_resolved_ids = rec["$output_id_mapping"]["resolved_ids"]
            input_id = self._extractInputID(input_resolved_ids, input_type)
            output_id = self._extractInputID(output_resolved_ids, output_type)
            if not input_id or not output_id:
                continue
            for s_id, o_id in zip(input_id, output_id):
                if input_type == "Gene":
                    s_id = "NCBIGene:" + str(s_id)
                if output_type == "Gene":
                    o_id = "NCBIGene:" + str(o_id)
                ngd_inputs.add(str(s_id) + "-" + str(o_id))
                id_dict[i].append(str(s_id) + "-" + str(o_id))
        annotatedResult = self._parseResponse(self._queryNGD(list(ngd_inputs)))
        cnt = 0
        for i, rec in enumerate(self.stepResult):
            if i in id_dict:
                ngd_scores = []
                for pair in id_dict[i]:
                    if pair in annotatedResult:
                        ngd_scores.append(annotatedResult[pair])
                if ngd_scores:
                    rec["$ngd"] = min(ngd_scores)
                    cnt += 1
        if cnt > 0:
            print(
                "Number of output edges sent to NGD score annotation is {}. Number of output edges annotated with NGD score is {}".format(
                    len(self.stepResult), cnt
                )
            )
        return

    def filter(self):
        res = []
        if "sort" in self.criteria:
            limit = (
                int(self.criteria["sort"].get("limit"))
                if self.criteria["sort"].get("limit")
                else 0
            )
            if self.criteria["sort"].get("direction") != "descend":
                idList = [item[0] for item in self.cnt.most_common(limit)]
                for rec in self.stepResult:
                    if rec["$output"] in idList:
                        res.append(rec)
                return res
            sortedList = self.cnt.most_common(len(self.cnt))
            sortedList.reverse()
            idList = [item[0] for i, item in enumerate(sortedList) if i < limit]
            for rec in self.stepResult:
                if rec["$output"] in idList:
                    res.append(rec)
            return res



import pandas as pd

from .utils import *
from ..smartapi_kg import MetaKG
from ..call_apis import APIQueryDispatcher
from ..config_new import BTE_FILTERS
from .filter.nodeDegree import NodeDegreeFilter
from .filter import Filter
from ..expand import Expander
from .printer import Print


class Predict:
    def __init__(self, input_objs, intermediate_nodes, output_types, config=None):
        self.input_objs = input_objs
        self.intermediate_nodes = intermediate_nodes
        if isinstance(intermediate_nodes, str):
            intermediate_nodes = [intermediate_nodes]
        self.output_types = output_types
        self.intermediate_nodes.append(self.output_types)
        validate_max_intermediate_nodes(self.intermediate_nodes)
        self.steps_results = {}
        self.steps_nodes = {}
        self.kg = MetaKG()
        self.kg.constructMetaKG(source="local")
        self.query_completes = False
        self.config = config
        self.ep = Expander()
        self.pt = Print()

    def _expand_inputs(self, inputs, verbose=True):
        """
        Expand inputs to its descendants.
        :param inputs: list of input bioentities with resolved ids.
        :param verbose: verbose
        """
        if (
            self.config
            and self.config.get("expand")
            and self.config.get("expand") is True
        ):
            if verbose:
                self.pt.print_expand_begin_message()
            expandedInputs = self.ep.expand(inputs.values())
            if verbose:
                self.pt.print_expand_summary_message(expandedInputs)
            if expandedInputs:
                inputs.update(expandedInputs)
        return inputs

    def _annotate_edges_with_filters(self, edges, step):
        """
        Add filter information to each edge.
        :param edges: list of edges.
        """
        if (
            self.config
            and self.config.get("filters")
            and isinstance(self.config["filters"], list)
            and step < len(self.config["filters"])
            and isinstance(self.config["filters"][step], dict)
            and self.config["filters"][step]
        ):
            if len(set(self.config["filters"][step].keys()) - set(BTE_FILTERS)) == 0:
                return
            for edge in edges:
                output_nodes = self.intermediate_nodes[step]
                if isinstance(output_nodes, str):
                    edge["filter"] = self.config["filters"][step]
                if isinstance(output_nodes, list):
                    for i, node in enumerate(output_nodes):
                        if node == edge["association"]["output_type"]:
                            if isinstance(self.config["filters"][step], list):
                                edge["filter"] = self.config["filters"][step][i]
                            else:
                                edge["filter"] = self.config["filters"][step]

    def _get_predicates_from_config(self, path):
        """
        Get information on predicates from config.
        """
        if (
            self.config
            and self.config.get("predicates")
            and isinstance(self.config["predicates"], list)
            and path < len(self.config["predicates"])
        ):
            predicates = self.config["predicates"][path]
        else:
            predicates = None
        return predicates

    def _annotate_results(self, step, source_types):
        # print("annotating results with NodeDegree!")
        # ft = NodeDegreeFilter(self.steps_results[step], {})
        # ft.annotateNodeDegree()
        # self.steps_results[step] = ft.stepResult
        if "annotate" in self.config and isinstance(self.config["annotate"], list):
            ft = Filter(self.steps_results[step], self.config["annotate"], source_types)
            ft.annotate()
            self.steps_results[step] = ft.stepResult
            return ft
        return None

    def _filter_results(self, step, ft=None):
        if (
            self.config
            and self.config.get("filters")
            and step < len(self.config["filters"])
        ):
            if not ft:
                ft = Filter(self.steps_results[step], self.config["filters"][step])
            else:
                ft.criteria = self.config["filters"][step]
            self.steps_results[step] = ft.filter_response()
            self.steps_nodes[step] = extractAllResolvedOutputIDs(
                self.steps_results[step]
            )
            print(
                "\nAfter applying post-query filter, BTE retrieved {} unique output nodes.".format(
                    len(self.steps_nodes[step])
                )
            )
        return self.steps_nodes[step]

    def connect(self, verbose=True):
        if not validate_max_intermediate_nodes(self.intermediate_nodes):
            return
        self.query_completes = False
        inputs = restructureHintOutput(self.input_objs)
        inputs = self._expand_inputs(inputs, verbose=verbose)
        if verbose:
            self.pt.print_query_parameter_summary(inputs, self.intermediate_nodes)
        for i, node in enumerate(self.intermediate_nodes):
            if verbose:
                self.pt.print_query_plan_begin_message(i + 1)
            inputs = groupsIDsbySemanticType(inputs)
            source_types = list(inputs.keys())
            predicates = self._get_predicates_from_config(i)
            if verbose:
                print(
                    "Input Types: {}\nOutput Types: {}\nPredicates: {}\n".format(
                        ",".join(list(inputs.keys())), node, predicates
                    )
                )
            edges = getEdges(inputs, node, predicates, self.kg)
            if len(edges) == 0:
                self.pt.print_query_failure_message("APIs", i + 1)
                return
            annotatedEdges = []
            for e in edges:
                annotatedEdges += annotateEdgesWithInput(
                    e.get("edges"), e.get("inputs")
                )
            if not annotatedEdges:
                self.pt.print_query_failure_message("APIs", i + 1)
                return
            self._annotate_edges_with_filters(annotatedEdges, i)
            dp = APIQueryDispatcher(annotatedEdges, verbose=verbose)
            self.steps_results[i] = dp.syncQuery()
            if not self.steps_results[i] or len(self.steps_results[i]) == 0:
                self.pt.print_query_failure_message("results", i + 1)
                return
            inputs = self.steps_nodes[i] = extractAllResolvedOutputIDs(
                self.steps_results[i]
            )
            if verbose:
                self.pt.print_individual_query_step_summary(inputs)
            ft = self._annotate_results(i, source_types)
            inputs = self._filter_results(i, ft)
        self.query_completes = True
        if verbose:
            self.pt.print_final_result_summary(self.steps_nodes)

    def annotate(self, step=0, criteria=[]):
        ft = Filter(self.steps_results[step], criteria=criteria)
        self.steps_results[step] = ft.annotate()

    def display_table_view(self, extra_fields=[]):
        if not self.query_completes:
            print("Your query fails. Unable to display results!")
            return

        for step, step_result in self.steps_results.items():
            df = stepResult2PandasTable(
                step_result, step, len(self.steps_results), extra_fields
            )
            if step == 0:
                result = df
                continue
            join_columns = [
                "node" + str(step) + item for item in ["_id", "_type", "_label"]
            ]
            result = pd.merge(result, df, on=join_columns, how="inner")
        return result



from functools import reduce
from collections import defaultdict

from ..utils.common import getCurieFromVal, dict_set2list
from .bioentity import BioEntity


class BioThingsAPI:
    MAX_BATCH_SIZE = 1000

    def __init__(self, metadata):
        self._process_metadata(metadata)

    def _process_metadata(self, metadata):
        self.url = metadata.get("url")
        self.mapping = metadata.get("mapping")
        self.rank = metadata.get("id_ranks")

    def _get_all_fields(self):
        return reduce(lambda x, y: x + y, self.mapping.values(), [])

    def _get_scope(self, prefix):
        return self.mapping.get(prefix)

    def build_query(self, prefix, ids):
        query = "q={inputs}&scopes={scopes}&fields={fields}&dotfield=true&species=human"
        for i in range(0, len(ids), self.MAX_BATCH_SIZE):
            yield (
                query.replace("{inputs}", ",".join(ids[i : i + self.MAX_BATCH_SIZE]))
                .replace("{scopes}", ",".join(self._get_scope(prefix)))
                .replace("{fields}", ",".join(self._get_all_fields()))
            )

    def __get_db_ids_helper(self, record):
        res = defaultdict(set)
        for k, v in self.mapping.items():
            for field_name in v:
                if field_name in record:
                    if isinstance(record[field_name], list):
                        for val in record[field_name]:
                            res[k].add(str(val))
                    else:
                        res[k].add(str(record[field_name]))
        return dict_set2list(res)

    def get_db_ids(self, prefix, semantic_type, response):
        result = {}
        for rec in response:
            if "notfound" not in rec:
                curie = getCurieFromVal(rec.get("query"), prefix)
                result[curie] = BioEntity(semantic_type, self.__get_db_ids_helper(rec))
        return result



# -*- coding: utf-8 -*-
"""
Restructure the output from OpenTarget API.

.. moduleauthor:: Jiwen Xin <kevinxin@scripps.edu>

"""


def restructure_opentarget_response(json_doc):
    """
    Restructure the API output from opentarget API.

    :param: json_doc: json output from opentarget API
    """
    if not json_doc.get("data"):
        return json_doc
    for _doc in json_doc['data']:
        if "drug" in _doc:
            if "CHEMBL" in _doc['drug']['id']:
                _doc['drug']['id'] = _doc['drug']['id'].split('/')[-1]
    return json_doc



# -*- coding: utf-8 -*-

"""
Make API calls.

.. moduleauthor:: Jiwen Xin <kevinxin@scripps.edu>

"""
from collections import defaultdict
from copy import deepcopy
from itertools import groupby
from operator import itemgetter
from .registry import Registry
from .apicall import BioThingsCaller
from .api_output_parser import OutputParser
from .api_preprocess import APIPreprocess

class Dispatcher:

    """Dispatch API calls."""

    def __init__(self, registry=None):
        """Load BTE registry and API caller."""
        if not registry:
            self.registry = Registry().registry
        else:
            self.registry = registry.registry
        self.caller = BioThingsCaller()

    @staticmethod
    def get_unique_edge_id(edge):
        operation = edge["operation"]
        _id = "-".join(
            [
                str(edge["value"]),
                operation["server"],
                operation["method"],
                operation["path"],
            ]
        )
        request_body = operation.get("requestBody")
        if request_body and request_body.get("body"):
            for k in sorted(request_body.get("body").keys()):
                _id += "-" + k + "-" + str(request_body.get("body")[k])
        parameters = operation.get("parameters")
        if parameters:
            for k in sorted(parameters.keys()):
                _id += "-" + k + "-" + str(parameters[k])
        return _id

    @staticmethod
    def get_all_edges(query_id2inputs_mapping):
        res = []
        for v in query_id2inputs_mapping.values():
            if isinstance(v, list) and len(v) > 0:
                res.append(v[0])
        return res

    @staticmethod
    def group_edges(edges):
        grouper = itemgetter("operation_id")
        new_edges = []
        # group all edges based on their API and input_field value
        for _, grp in groupby(sorted(edges, key=grouper), grouper):
            grp = list(grp)
            new_edge = deepcopy(grp[0])
            values = set()
            for edge in grp:
                if isinstance(edge["value"], list):
                    values |= set(edge["value"])
                else:
                    values.add(edge["value"])
            new_edge["value"] = values
            new_edges.append(new_edge)
        return new_edges

    @staticmethod
    def subset_mapping_file(edges, mapping_file):
        """Only maintain a subset of mapping file based on edge label."""
        mapping_keys = [_item.get("mapping_key") for _item in edges]
        if mapping_keys:
            mapping_keys += ["@type", "@context"]
            return {k: v for (k, v) in mapping_file.items() if k in mapping_keys}
        return mapping_file

    def construct_api_calls(self, edges):
        """Construct API calls for apicall module using edge groups."""
        unique_edge_ids = set()
        edge_id2query_id_mapping = {}
        query_id2inputs_mapping = defaultdict(list)
        edges = self.group_edges(edges)
        for edge in edges:
            api = edge["api"]
            if edge["operation"].get("supportBatch"):
                edge["value"] = edge["operation"]["inputSeparator"].join(edge["value"])
                edge_id = self.get_unique_edge_id(edge)
                if edge_id in unique_edge_ids:
                    internal_query_id = edge_id2query_id_mapping[edge_id]
                else:
                    internal_query_id = (
                        "API "
                        + self.api_dict[api]["num"]
                        + "."
                        + str(self.api_dict[api]["alphas"].pop(0))
                    )
                    edge["internal_query_id"] = internal_query_id
                    edge_id2query_id_mapping[edge_id] = internal_query_id
                    unique_edge_ids.add(edge_id)
                edge["internal_query_id"] = internal_query_id
                query_id2inputs_mapping[internal_query_id].append(edge)
                # internal_query_id += 1
            else:
                for val in edge["value"]:
                    new_edge = deepcopy(edge)
                    new_edge["value"] = val
                    edge_id = self.get_unique_edge_id(new_edge)
                    if edge_id in unique_edge_ids:
                        internal_query_id = edge_id2query_id_mapping[edge_id]
                    else:
                        internal_query_id = (
                            "API "
                            + self.api_dict[api]["num"]
                            + "."
                            + str(self.api_dict[api]["alphas"].pop(0))
                        )
                        edge_id2query_id_mapping[edge_id] = internal_query_id
                        unique_edge_ids.add(edge_id)
                    new_edge["internal_query_id"] = internal_query_id
                    query_id2inputs_mapping[internal_query_id].append(new_edge)
                    # internal_query_id += 1
        return query_id2inputs_mapping

    @staticmethod
    def add_metadata_to_output(operation, res, output_id):
        if isinstance(res, dict):
            if output_id not in res:
                return []
            if not isinstance(res[output_id], list):
                res[output_id] = [res[output_id]]
            new_res = []
            for val in res[output_id]:
                tmp = deepcopy(res)
                tmp[output_id] = val
                tmp.update(
                    {
                        "$api": operation["api_name"],
                        "$source": operation.get("source"),
                        "@type": operation["output_type"],
                    }
                )
                new_res.append(tmp)
            return new_res
        res = {
            "$api": operation["api_name"],
            "$source": operation.get("source"),
            "@type": operation["output_type"],
            operation["output_id"]: [res],
        }
        return [res]

    @staticmethod
    def count_hits(res):
        cnt = 0
        if not res:
            return cnt
        for pred_infos in res.values():
            if pred_infos:
                for info in pred_infos.values():
                    cnt += len(info)
        return cnt

    def dispatch(self, edges, verbose=False, loop=None):
        """Send request to and parse response from API."""
        results = {}
        self.unique_apis = {_edge["api"] for _edge in edges if _edge}
        self.api_dict = {}
        for i, _api in enumerate(list(self.unique_apis)):
            self.api_dict[_api] = {"alphas": list(range(1, 10000)), "num": str(i + 1)}
        query_id2inputs_mapping = self.construct_api_calls(edges)
        # print(apis, inputs, modes, vals, grped_edges)
        responses, self.log = self.caller.call_apis(
            self.get_all_edges(query_id2inputs_mapping), verbose=verbose, loop=loop
        )
        if verbose:
            print("\n\n==== Step #3: Output normalization ====\n")
        self.log.append("\n\n==== Step #3: Output normalization ====\n")
        for response in responses:
            if response["result"] == {}:
                continue
            output_types = []
            query_id = response["internal_query_id"]
            total_hits = 0
            for edge in query_id2inputs_mapping[query_id]:
                operation = edge["operation"]
                api_name = edge["api"]
                if api_name[:4] in ["semm", "cord"]:
                    output_types = [edge["output_type"]]
                _res = APIPreprocess(
                    response["result"], operation["api_type"], api_name, output_types
                ).restructure()
                mapping = operation["response_mapping"]
                _res = OutputParser(
                    _res,
                    mapping,
                    operation["supportBatch"],
                    api_name,
                    operation["api_type"],
                ).parse()
                if not operation["supportBatch"]:
                    # preprocess biolink results
                    # if val is not present in results dict and _res is not empty
                    if not _res:
                        continue
                    val = edge["value"]
                    if val not in results:
                        results[val] = {}
                    # loop through API call response
                    for k, v in _res.items():
                        # if key is not present in final res, create a list
                        if k not in results[val]:
                            results[val][k] = []
                        if isinstance(v, list):
                            for _v in v:
                                _v = self.add_metadata_to_output(
                                    operation, _v, operation["output_id"]
                                )
                                total_hits += len(_v)
                                results[val][k] += _v
                        else:
                            v = self.add_metadata_to_output(
                                operation, v, operation["output_id"]
                            )
                            total_hits += len(v)
                            results[val][k] += v
                else:
                    if not _res:
                        continue
                    for m, n in _res.items():
                        if m not in results:
                            results[m] = {}
                        for k, v in n.items():
                            if k not in results[m]:
                                results[m][k] = []
                            if isinstance(v, list):
                                for _v in v:
                                    _v = self.add_metadata_to_output(
                                        operation, _v, operation["output_id"]
                                    )
                                    total_hits += len(_v)
                                    results[m][k] += _v
                            elif isinstance(v, dict):
                                v = self.add_metadata_to_output(
                                    operation, v, operation["output_id"]
                                )
                                total_hits += len(v)
                                results[m][k] += v
                            else:
                                if k == "query":
                                    continue
                                v = self.add_metadata_to_output(
                                    operation, v, operation["output_id"]
                                )
                                total_hits += len(v)
                                results[m][k] += v
            if verbose:
                if total_hits > 0:
                    print("{} {}: {} hits".format(query_id, api_name, total_hits))
                else:
                    print("{} {}: No hits".format(query_id, api_name))
            if total_hits > 0:
                self.log.append("{} {}: {} hits".format(query_id, api_name, total_hits))
            else:
                self.log.append("{} {}: No hits".format(query_id, api_name))
        return (dict(results), self.log)



"""Generate a json file containing semmed relationships

The python script takes two arguments:
1. DATA_FOLDER: the folder storing neo4j_nodes.csv and neo4j_edges.csv files
2. OUTPUT_FILE_PATH: the file path storing the output
"""
from collections import defaultdict
import csv
import os
import sys
import json

maxInt = sys.maxsize


SEMMED_SEMANTIC_TYPE_MAPPING = {
    "chemical_substance": "ChemicalSubstance",
    "activity_and_behavior": None,
    "anatomical_entity": "AnatomicalEntity",
    "biological_entity": None,
    "biological_process_or_activity": "BiologicalProcess",
    "cell": "Cell",
    "cell_component": "CellularComponent",
    "disease_or_phenotypic_feature": "DiseaseOrPhenotypicFeature",
    "gene": "Gene",
    "genomic_entity": None,
    "gross_anatomical_structure": None,
    "phenotypic_feature": "PhenotypicFeature",
    "protein": "Gene"
}
SEMMED_PRED_MAPPING = {
    "ASSOCIATED_WITH": {
        "self": "related_to",
        "reverse": "related_to"
    },
    "INTERACTS_WITH": {
        "self": "physically_interacts_with",
        "reverse": "physically_interacts_with"
    },
    "AFFECTS": {
        "self": "affects",
        "reverse": "affected_by"
    },
    "STIMULATES": {
        "self": "positively_regulates",
        "reverse": "positively_regulated_by"
    },
    "INHIBITS": {
        "self": "negatively_regulates",
        "reverse": "negatively_regulated_by"
    },
    "DISRUPTS": {
        "self": "disrupts",
        "reverse": "disrupted_by"
    },
    "COEXISTS_WITH": {
        "self": "coexists_with",
        "reverse": "coexists_with"
    },
    "PREDISPOSES": {
        "self": "predisposes",
        "reverse": "predisposed_by"
    },
    "CAUSES": {
        "self": "causes",
        "reverse": "caused_by"
    },
    "TREATS": {
        "self": "treats",
        "reverse": "treated_by"
    },
    "PREVENTS": {
        "self": "prevents",
        "reverse": "prevented_by"
    },
    "OCCURS_IN": {
        "self": "occurs_in",
        "reverse": "in_which_occured"
    },
    "PROCESS_OF": {
        "self": "occurs_in",
        "reverse": "in_which_occured"
    },
    "LOCATION_OF": {
        "self": "location_of",
        "reverse": "located_in"
    },
    "PART_OF": {
        "self": "part_of",
        "reverse": "has_part"
    },
    "USES": {
        "self": "has_input",
        "reverse": "is_input_of"
    },
    "CONVERTS_TO": {
        "self": "derives_info",
        "reverse": "derives_from"
    },
    "MANIFESTATION_OF": {
        "self": "manifestation_of",
        "reverse": "manifested_by"
    },
    "PRODUCES": {
        "self": "produces",
        "reverse": "produced_by",
    },
    "PRECEDES": {
        "self": "precedes",
        "reverse": "preceded_by"
    },
    "ISA": {
        "self": "subclass_of",
        "reverse": "has_subclass"
    }
}

while True:
    # decrease the maxInt value by factor 10 
    # as long as the OverflowError occurs.

    try:
        csv.field_size_limit(maxInt)
        break
    except OverflowError:
        maxInt = int(maxInt/10)


def load_data(data_folder):
    edges_dict = {v: defaultdict(set) for k, v in SEMMED_SEMANTIC_TYPE_MAPPING.items() if v != None}
    nodes_path = os.path.join(data_folder, "nodes_neo4j.csv")
    edges_path = os.path.join(data_folder, "edges_neo4j.csv")
    group_by_semmantic_dict = defaultdict(list)
    id_type_mapping = {}
    with open(nodes_path) as f:
        csv_reader = csv.reader(f, delimiter=',')
        next(csv_reader)
        for _item in csv_reader:
            group_by_semmantic_dict[_item[-2]].append(_item[-1])
            id_type_mapping[_item[-1]] = {'type': _item[-2], 'name': _item[1]}
    gene_related = {}
    with open(edges_path) as f:
        csv_reader = csv.reader(f, delimiter=',')
        next(csv_reader)
        for _item in csv_reader:
            source_type = SEMMED_SEMANTIC_TYPE_MAPPING[id_type_mapping[_item[4]]['type']]
            pred = SEMMED_PRED_MAPPING[_item[0]]['self']
            pred_reverse = SEMMED_PRED_MAPPING[_item[0]]['reverse']
            target_type = SEMMED_SEMANTIC_TYPE_MAPPING[id_type_mapping[_item[5]]['type']]
            if source_type in edges_dict and target_type in edges_dict:
                edges_dict[source_type][pred].add(target_type)
                edges_dict[target_type][pred_reverse].add(source_type)
    for k, v in edges_dict.items():
        for m, n in v.items():
             v[m] = list(n)
    return dict(edges_dict)


if __name__ == '__main__':
    DATA_FOLDER = sys.argv[1]
    print("DATA_FOLDER", DATA_FOLDER)
    OUTPUT_FILE_PATH = sys.argv[2]
    print("OUTPUT_FILE_PATH", OUTPUT_FILE_PATH)

    a = load_data(DATA_FOLDER)
    with open(OUTPUT_FILE_PATH, 'w') as outfile:
        json.dump(a, outfile, indent=4, sort_keys=True)



from copy import deepcopy

def restructure_cord_response(json_doc, output_types):
    """Restructure the JSON output from cord API.

    :param: json_doc: the API response from cord API
    """
    if not isinstance(json_doc, list):
        return json_doc
    new_res = []
    for _res in json_doc:
        if not isinstance(_res, dict):
            continue
        # handle case where the queried item is not found
        if _res.get('notfound'):
            continue
        tmp_res = {'query': _res['query']}
        for k, v in _res.items():
            if k == "associated_with":
                k = "related_to"
            tmp_v = []
            if isinstance(v, list):
                for item in v:
                    if item["@type"] in output_types or (item["@type"] == 'DiseaseOrPhenotypicFeature' and output_types == ['Disease']):
                        item_copy = deepcopy(item)
                        for m in item.keys():
                            if m in ['pr', 'go', 'mop', 'hgnc', 'uberon', 'so', 'cl', 'doid', 'chebi']:
                                item_copy[m.upper()] = item_copy.pop(m)
                        tmp_v.append(item_copy)
            if tmp_v != []:
                tmp_res[k] = tmp_v
        new_res.append(tmp_res)
    return new_res


# -*- coding: utf-8 -*-
"""
Export the output of BTE to graphviz.

.. moduleauthor:: Jiwen Xin <kevinxin@scripps.edu>


"""
from graphviz import Digraph


def visualize(edges, size=None):
    """Export networkx graph into graphviz format for jupyter notebook visualization.

    :param: edges: the edge representation of the graph
    :param: size: the size of the output graph, e.g. (6, 8)
    """
    if size:
        d = Digraph(graph_attr=[('size', size)])    # pylint: disable=undefined-variable
    else:
        d = Digraph()                               # pylint: disable=undefined-variable
    for _item in edges:
        d.edge(_item[0], _item[1])
    return d

def networkx2graphvis(G):
    f = Digraph()
    for k, v, j in G.edges(data=True):
        f.edge(k, v, j['label'])
    return f


from collections import defaultdict

from ..smartapi_kg import MetaKG
from ..call_apis import APIQueryDispatcher
from ..query.utils import annotateEdgesWithInput


class Expander:
    def __init__(self):
        self.kg = MetaKG()
        self.kg.constructMetaKG(source="local")

    def __getEdges(self, semanticType):
        """
        Get a list of smart-api edges based on semantic type.

        :param semanticType: Type of bioentities to expand
        :returns: list of smartapi edges expanding the semantic type
        """
        return self.kg.filter(
            {
                "input_type": semanticType,
                "output_type": semanticType,
                "predicate": "has_subclass",
            }
        )

    @staticmethod
    def __parseResponse(res):
        if not res:
            return
        result = {}
        for rec in res:
            if (
                "$output_id_mapping" in rec
                and "resolved_ids" in rec["$output_id_mapping"]
            ):
                result[
                    rec["$output_id_mapping"]["resolved_ids"]["id"]["identifier"]
                ] = rec["$output_id_mapping"]["resolved_ids"]
        return result

    @staticmethod
    def __groupIDsbySemanticType(output_ids):
        result = defaultdict(list)
        for resolved_ids in output_ids:
            result[resolved_ids.get("type")].append(resolved_ids)
        return result

    def expand(self, inputs):
        """
        Expand input biomedical objects to its children
        :param semanticType: semantic type of the inputs
        :param inputs: list of resolved identifiers
        """
        grpedIDs = self.__groupIDsbySemanticType(inputs)
        bte_edges = []
        for semanticType, resolvedIDs in grpedIDs.items():
            smartapi_edges = self.__getEdges(semanticType)
            if not smartapi_edges:
                continue
            tmp_edges = annotateEdgesWithInput(smartapi_edges, resolvedIDs)
            if not tmp_edges:
                continue
            bte_edges += tmp_edges
        if not bte_edges:
            return
        dp = APIQueryDispatcher(bte_edges)
        res = dp.syncQuery()
        return self.__parseResponse(res)



# -*- coding: utf-8 -*-
"""
Resolving Biomedical Identifiers through BioThings APIs.

.. moduleauthor:: Jiwen Xin <kevinxin@scripps.edu>
"""

from .config_new import ID_RESOLVING_APIS
from .apicall import BioThingsCaller


class Hint:
    """Resolving Biomedical Identifiers through BioThings APIs."""

    def __init__(self):
        """Load BTE API caller."""
        self.caller = BioThingsCaller()

    def get_primary_id(self, semantic_type, json_doc):
        """Get the primary id of a biological entity
        
        :param: json_doc: the API output
        :param: semantic_type: the main entity type of the output, e.g. Gene, SequenceVariant
        """
        # parse the id rank info from metadata
        ranks = self.ID_RESOLVING_APIS[semantic_type]["id_ranks"]
        res = {}
        # loop through the id rank list, e.g. ['chembl', 'drugbank', ...]
        # the id rank list is ranked based on priorirty
        for _id in ranks:
            # if an id of higher priority is found, set it as the primary id
            if _id in json_doc:
                res["identifier"] = _id
                res["cls"] = semantic_type
                res["value"] = json_doc[_id]
                break
        return res

    def get_display_message(self, semantic_type, json_doc):
        """Get the primary id of a biological entity
        
        :param: semantic_type: the main entity type of the output, e.g. Gene, SequenceVariant
        :param: json_doc: the output
        """
        # parse the id rank info from metadata
        res = ""
        for _id in self.ID_RESOLVING_APIS[semantic_type]["id_ranks"]:
            if _id in json_doc:
                res += _id + "(" + str(json_doc[_id]) + ") "
        return res.strip(" ")

    @staticmethod
    def get_query_fields(mapping_file):
        """Parse the mapping file to retrieve fields for query"""
        fields = []
        for v in mapping_file.values():
            fields += v
        return ",".join(fields)

    def query(self, _input, semantic_type=None):
        """Main function to resolve identifiers.
        
        :param: _input: any string input
        """
        if semantic_type:
            if semantic_type not in ID_RESOLVING_APIS:
                raise ValueError(
                    "{} is not a valid semantic type to resolve. Available semantic types are {}".format(
                        semantic_type, ",".join(ID_RESOLVING_APIS.keys())
                    )
                )
            self.ID_RESOLVING_APIS = {semantic_type: ID_RESOLVING_APIS[semantic_type]}
        else:
            self.ID_RESOLVING_APIS = ID_RESOLVING_APIS
        self.construct_api_calls(_input)
        # make API calls asynchronously and gather all outputs
        self.responses, _ = self.caller.call_apis(self.api_call_inputs)
        return self.parse_api_responses()

    def parse_api_responses(self):
        """Parse the API responses from APICall module in BTE."""
        # bp,  pathway,  cc, mf shares the same API for id resolving
        TYPE_MAPPING = {
            "bp": "BiologicalProcess",
            "pathway": "Pathway",
            "cc": "CellularComponent",
            "mf": "MolecularActivity",
        }
        parsed_response = {k: [] for k in self.ID_RESOLVING_APIS}
        for _res in self.responses:
            # if API response is empty, continue
            if not _res:
                continue
            semantic_type = _res.get("internal_query_id")
            if not _res.get("result"):
                continue
            for single_res in _res["result"]:
                if single_res.get("notfound"):
                    continue
                tmp = {}
                if single_res.get("type") in TYPE_MAPPING.keys():
                    semantic_type = TYPE_MAPPING[single_res.get("type")]
                mapping = ID_RESOLVING_APIS[semantic_type]["mapping"]
                for k, v in mapping.items():
                    for _v in v:
                        if _v in single_res:
                            val = single_res[_v]
                            if not isinstance(val, list):
                                tmp[k] = val
                            elif len(val) > 0:
                                tmp[k] = val[0]
                            break
                tmp["primary"] = self.get_primary_id(semantic_type, tmp)
                tmp["display"] = self.get_display_message(semantic_type, tmp)
                tmp["type"] = semantic_type
                parsed_response[semantic_type].append(tmp)

        return parsed_response

    def construct_api_calls(self, _id):
        """Construct API calls for BTE API call module.
        """
        self.api_call_inputs = []
        unique_base_urls = set()
        for semantic_type, api_data in self.ID_RESOLVING_APIS.items():
            server_url = api_data.get("url")
            if server_url in unique_base_urls:
                continue
            unique_base_urls.add(server_url)
            query_id = semantic_type
            query_fields = self.get_query_fields(api_data["mapping"])
            if api_data["api_name"] == "geneset API":
                query_id = api_data["api_name"]
                query_fields = "go,umls,name,reactome,wikipathways,kegg,pharmgkb,type"
            self.api_call_inputs.append(
                {
                    "api": api_data["api_name"],
                    "operation": {
                        "server": server_url,
                        "path": "/query",
                        "method": "post",
                        "parameters": {
                            "fields": query_fields,
                            "dotfield": "true",
                            "species": "human",
                            "size": "5",
                        },
                        "requestBody": {
                            "body": {"q": '"{inputs[0]}"', "scopes": query_fields},
                            "header": "application/x-www-form-urlencoded",
                        },
                    },
                    "value": _id,
                    "internal_query_id": query_id,
                }
            )



# -*- coding: utf-8 -*-
"""Module to make asynchronous API calls.

.. moduleauthor:: Jiwen Xin <kevinxin@scripps.edu>


"""

import asyncio
from aiohttp import ClientSession, TCPConnector
from collections import Counter
import json
from .utils.common import add_s
import requests



class BioThingsCaller:
    """Call biothings APIs."""

    @staticmethod
    def print_request(method, url, params, request_body):
        if params:
            url += "?"
            params = [(str(m) + "=" + str(n)) for m, n in params.items()]
            url += "&".join(params)
        if method == "get":
            return url
        if method == "post":
            if request_body:
                url += " (POST -d "
                request_body = [
                    (str(m) + "=" + str(n)) for m, n in request_body.items()
                ]
                url += "&".join(request_body)
                url += ")"
            return url
        return ""

    async def call_one_arbitrary_api(self, _input, session, verbose=False):
        base_url = (
            _input["operation"]["server"].strip("/") + _input["operation"]["path"]
        )
        method = _input["operation"]["method"]
        request_body = _input["operation"].get("requestBody")
        header = {"content-type": "application/x-www-form-urlencoded"}
        if request_body:
            if request_body.get("header"):
                header = {"content-type": request_body.get("header")}
            request_body = eval(
                str(request_body["body"]).replace("{inputs[0]}", _input["value"])
            )
        parameters = _input["operation"].get("parameters")
        if parameters:
            if _input["operation"].get("path_params"):
                for path_param in _input["operation"]["path_params"]:
                    path_value_template = parameters.get(path_param)
                    # if(path_value_template):
                    base_url = base_url.replace(
                        "{" + path_param + "}", path_value_template
                    ).replace("{inputs[0]}", _input["value"])
                    parameters.pop(path_param)
            parameters = eval(str(parameters).replace("{inputs[0]}", _input["value"]))
        query_url = self.print_request(method, base_url, parameters, request_body)
        if method == "get":
            try:
                async with session.get(base_url, params=parameters) as res:
                    if verbose:
                        print("{}: {}".format(_input["internal_query_id"], query_url))
                    try:
                        if res.status in [400, 404]:
                            print(
                                "{} {} failed".format(
                                    _input["internal_query_id"], _input["api"]
                                )
                            )
                            return {
                                "internal_query_id": _input["internal_query_id"],
                                "result": {},
                            }
                        res = await res.json()
                        return {
                            "internal_query_id": _input["internal_query_id"],
                            "result": res,
                        }
                    except Exception as ex:
                        m = await res.text()
                        return {
                            "result": json.loads(m),
                            "internal_query_id": _input["internal_query_id"],
                        }
            except Exception:
                if verbose:
                    print(
                        "{} {} failed".format(
                            _input["internal_query_id"], _input["api"]
                        )
                    )
                return {"internal_query_id": _input["internal_query_id"], "result": {}}
        elif method == "post":
            if("mychem.info" not in base_url):
                # execute asynchronous calls as per usual
                try:
                    async with session.post(
                        base_url, params=parameters, data=request_body, headers=header
                    ) as res:
                        try:
                            if res.status in [400, 404]:
                                print(
                                    "{} {} failed".format(
                                        _input["internal_query_id"], _input["api"]
                                    )
                                )
                                return {
                                    "internal_query_id": _input["internal_query_id"],
                                    "result": {},
                                }
                            if verbose:
                                print(
                                    "{}: {}".format(_input["internal_query_id"], query_url)
                                )
                            return {
                                "result": await res.json(),
                                "internal_query_id": _input["internal_query_id"],
                            }
                        except Exception as ex1:
                            print(ex1)
                            print("Unable to fetch results from {}".format(_input["api"]))
                            return {
                                "internal_query_id": _input["internal_query_id"],
                                "result": {},
                            }
                except Exception as ex:
                    print(ex)
                    if verbose:
                        print(
                            "{}: {} failed".format(
                                _input["internal_query_id"], _input["api"]
                            )
                        )
                    return {"result": {}, "internal_query_id": _input["internal_query_id"]}
            else:
                # in this case, the call is to MyChem.info
                counter = 0
                interval = 500
                res = []
                request_list = request_body["q"].split(",")
                # only make queries with up to 200 items at a time - too many will return error
                while(counter < len(request_list)):
                    s = ","
                    request_body["q"] = s.join(request_list[counter:(counter+interval)])
                    # make synchronous calls
                    res_temp = requests.post(base_url, params=parameters, data=request_body, headers=header)
                    if(res_temp.status_code == 200):
                        # combine responses from 1+ calls
                        res = res + res_temp.json()
                        counter = counter + interval
                try:
                    return {
                        "result": res,
                        "internal_query_id": _input["internal_query_id"]
                    }
                except Exception as ex2:
                    print(ex2)
                    print("Unable to fetch results from {}".format(_input["api"]))
                    return {
                        "internal_query_id": _input["internal_query_id"],
                        "result": {},
                    }

    async def call_one_api(self, _input, session, verbose=False):
        """Asynchronously make one API call.

        :param: _input (dict) : a python dict containing three keys, e.g. batch_mode, params, api
        :param: session (obj): a aiohttp session object
        """
        # check api type
        res = await self.call_one_arbitrary_api(_input, session, verbose=verbose)
        return res

    async def run(self, inputs, verbose=False):
        """Asynchronously make a list of API calls.

        :param: inputs (list): list of python dicts containing three keys
        """
        tasks = []
        # timeout = ClientTimeout(total=15)
        async with ClientSession(connector=TCPConnector(ssl=False)) as session:
            for i in inputs:
                task = self.call_one_api(i, session, verbose=verbose)
                tasks.append(task)
            responses = await asyncio.gather(*tasks)
            # print(responses)
            return responses

    def call_apis(self, inputs, verbose=False, loop=None):

        self.log = []
        if verbose:
            cnt = Counter()
            if inputs:
                for _input in inputs:
                    cnt[_input["api"]] += 1
            self.unique_apis = {_edge["api"] for _edge in inputs if _edge}
            if verbose:
                print("\nBTE found {} apis:\n".format(len(self.unique_apis)))
                for i, _api in enumerate(self.unique_apis):
                    print(
                        "API {}. {}({} API call{})".format(
                            i + 1, _api, cnt[_api], add_s(cnt[_api])
                        )
                    )
            self.log.append("\nBTE found {} apis:\n".format(len(self.unique_apis)))
            for i, _api in enumerate(self.unique_apis):
                self.log.append(
                    "API {}. {}({} API call{})".format(
                        i + 1, _api, cnt[_api], add_s(cnt[_api])
                    )
                )
            print("\n\n==== Step #2: Query path execution ====")
            self.log.append("\n\n==== Step #2: Query path execution ====")
            print(
                "NOTE: API requests are dispatched in parallel, so the list of APIs below is ordered by query time.\n"
            )
            self.log.append(
                "NOTE: API requests are dispatched in parallel, so the list of APIs below is ordered by query time.\n"
            )
        if not loop:
            loop = asyncio.new_event_loop()
        return (loop.run_until_complete(self.run(inputs, verbose=verbose)), self.log)
        # res = asyncio.run(self.run(inputs, verbose=verbose))
        # return (res, self.log)



# -*- coding: utf-8 -*-

"""
biothings_explorer.mapping_parser
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module contains code which parses the mapping file between
biothings schema and biothings API fields
"""
from collections import defaultdict
import itertools

import networkx as nx
from .schema_parser import SchemaParser
from .utils.dataload import load_json_or_yaml
from .utils.common import find_longest_common_path, get_dict_values, remove_prefix
from .utils.cord import cord, SEMANTIC_TYPE_ID_MAPPING
from .utils.semmed import semmed
from .config import metadata, PREFIX_TO_REMOVE


class MappingParser():
    """Parse the mapping file between biothings schema and biothings API"""
    BIOTHINGS_SCHEMA_PATH = 'https://raw.githubusercontent.com/data2health/schemas/biothings/biothings/biothings_curie_kevin.jsonld'

    def __init__(self):
        self.se = SchemaParser()
        # list all properties which are descendants of identifiers
        self.id_list = self.se.ids
        # get all classes defined in biothings schema JSON-LD file
        self.defined_clses = self.se.clses
        # list of properties whose "range" is among defined classes
        self.linked_prop_list = self.se.properties.keys()
        self.cls_prop_clsf = {}

    def load_mapping(self, mapping, api=None):
        self.mapping = load_json_or_yaml(mapping)
        self.mapping = remove_prefix(self.mapping, PREFIX_TO_REMOVE)
        self.api = api

    def classify_keys_in_json(self, json_doc):
        """ classify the keys in a json doc"""
        result = defaultdict(list)
        for _key in json_doc.keys():
            if _key in self.id_list:
                result['id'].append(_key)
            elif _key in self.linked_prop_list:
                result['links'].append(_key)
        return result

    def connect(self):
        # G is used to store the relationship between input and output in a schema mapping file
        G = nx.MultiDiGraph()
        # get the document type
        self.type = self.mapping.get("@type")
        if metadata[self.api].get('api_name') == 'CORD API':
            for pred, info in cord[self.type].items():
                for output_type in info:
                    for input_id in SEMANTIC_TYPE_ID_MAPPING[self.type]:
                        for output_id in SEMANTIC_TYPE_ID_MAPPING[output_type]:
                            G.add_edge(
                                input_id.lower(), output_id.lower(),
                                label=pred,
                                mapping_key=pred,
                                api=self.api,
                                source='CORD API',
                                input_field=input_id.lower(),
                                input_id=input_id.lower(),
                                input_type=self.type,
                                output_id=output_id.lower(),
                                output_field=pred,
                                output_type=output_type
                            )
        elif metadata[self.api].get('api_name') == 'SEMMED API':
            for pred, info in semmed[self.type].items():
                for output_type in info:
                    G.add_edge(
                        'umls', 'umls',
                        label=pred,
                        mapping_key=pred,
                        api=self.api,
                        source='SEMMED API',
                        input_field='umls',
                        input_id='umls',
                        input_type=self.type,
                        output_id='umls',
                        output_field=pred,
                        output_type=output_type
                    )
        else:
            # classify the keys in the JSON doc into IDs or Links
            clsf = self.classify_keys_in_json(self.mapping)
            # for each "links" properties, find its ids
            for predicate in clsf['links']:
                if isinstance(self.mapping[predicate], dict):
                    self.mapping[predicate] = [self.mapping[predicate]]
                for _pred in self.mapping[predicate]:
                    if "@type" in _pred:
                        sp = self.se.properties.get(predicate)
                        obj_clsf = self.classify_keys_in_json(_pred)
                        common_prefix = find_longest_common_path(get_dict_values(_pred))
                        input_id = [_pred['$input']] if '$input' in _pred else clsf['id']
                        source = _pred['$source'] if '$source' in _pred else self.api
                        for _edge in itertools.product(input_id, obj_clsf['id']):
                            output_field = _pred[_edge[1]]
                            input_field = self.mapping[_edge[0]]
                            if isinstance(input_field, list):
                                input_field = ','.join(input_field)
                            if isinstance(output_field, list):
                                output_field = ','.join(output_field)
                            G.add_edge(_edge[0], _edge[1], label=predicate,
                                    mapping_key=predicate,
                                    api=self.api,
                                    source=source,
                                    input_field=input_field,
                                    input_type=self.mapping["@type"],
                                    input_id=_edge[0],
                                    output_id=_edge[1],
                                    output_type=_pred["@type"],
                                    output_field=common_prefix if common_prefix else output_field)
                            if metadata[self.api].get('api_type') == 'biothings':
                                inverse_property = sp.get('inverse_property')
                                if not inverse_property:
                                    print(predicate)
                                G.add_edge(_edge[1], _edge[0],
                                            api=self.api,
                                            input_field=output_field,
                                            input_type=_pred["@type"],
                                            source=source,
                                            input_id=_edge[1],
                                            output_id=_edge[0],
                                            output_type=self.mapping["@type"],
                                            output_field=input_field,
                                            label=inverse_property,
                                            mapping_key=_edge[0])
        return G



from .biothings_transformer import BioThingsTransformer


class DrugResponseTransformer(BioThingsTransformer):
    def wrap(self, res):
        if "filter" not in self.edge or "disease_context" not in self.edge["filter"]:
            return []
        if "hits" not in res or not isinstance(res.get("hits"), list):
            return res
        for doc in res.get("hits"):
            if not doc.get("association"):
                continue
            if not doc["association"].get("effect_size"):
                continue
            if not doc["association"].get("pvalue"):
                continue
            doc["association"]["pvalue"] = str(doc["association"]["pvalue"])
            doc["association"]["effect_size"] = str(doc["association"]["effect_size"])
        return res



from setuptools import setup, find_packages


install_requires = [
    "jupyter",
    "ipykernel",
    "ipython",
    "notebook",
    "networkx==2.4",
    "jsonpath-rw>=1.4.0",
    "requests>=2.21.0",
    "graphviz>=0.11.1",
    "aiohttp",
    "pandas",
    "pyyaml",
    "nest_asyncio",
]

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="biothings_explorer",
    version="1.0.2",
    author="Jiwen Xin, Chunlei Wu",
    author_email="cwu@scripps.edu",
    description="Python Client for BioThings Explorer",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license="BSD",
    keywords="schema biothings",
    url="https://github.com/biothings/biothings_explorer",
    packages=find_packages(),
    include_package_data=True,
    package_data={"": ["data/smartapi_local_specs.json"]},
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Topic :: Utilities",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    install_requires=install_requires,
    dependency_links=["git+https://github.com/mmayers12/data_tools.git#egg=data_tools"],
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
)



import requests
from copy import deepcopy
import json
import os
import pkg_resources

from .config import SMARTAPI_URL


def restructure_specs(spec):
    """SmartAPI API restructure each spec when storing them. This function converts them back to the original smartapi specification"""
    copy_spec = deepcopy(spec)
    new_paths = {}
    if isinstance(copy_spec.get("paths"), list) and len(copy_spec.get("paths")) > 0:
        for path in copy_spec.get("paths"):
            new_paths[path.get("path")] = path.get("pathitem")
    copy_spec["paths"] = new_paths
    return copy_spec


def load_specs(source="remote", tag="translator"):
    """Load SmartAPI specs."""
    if source == "remote":
        try:
            specs = requests.get(SMARTAPI_URL).json()
        except:
            raise Exception(
                "Unable to retrieve smartapi specs from {}".format(SMARTAPI_URL)
            )
    else:
        DATA_PATH = pkg_resources.resource_filename("biothings_explorer", "data/")
        DB_FILE = pkg_resources.resource_filename(
            "biothings_explorer", "data/smartapi_local_specs.json"
        )
        dir_path = os.path.dirname(os.path.abspath(__file__))
        file_path = os.path.join(dir_path, "smartapi_local_specs.json")
        with open(DB_FILE) as f:
            specs = json.load(f)
    new_specs = []
    for spec in specs.get("hits"):
        tags = [item.get("name") for item in spec.get("tags")]
        if tag in tags:
            new_specs.append(restructure_specs(spec))
    return new_specs




"""Some common utils functions for BioThings Explorer"""

from os.path import commonprefix
from ..config import id_ranks, INTERNAL_KEYS
from ..config_new import ALWAYS_PREFIXED


def add_s(num: int) -> str:
    """Add 's' if num is more than one.

    :param: num: An integer representing count
    """
    if not isinstance(num, int):
        return ""
    if num <= 1:
        return ""
    return "s"


def dict2listoftuples(py_dict: dict) -> list:
    """Convert a single python dictionary into a list of tuples.

    :param: py_dict: a single python dictionary
    """
    return [(k, v) for k, v in py_dict.items()]


def listoftuples2dict(tuple_list: list) -> dict:
    """Convert a list of tuples back to a single dictionary.

    :param: tuple_list: a list of tuples
    """
    try:
        return dict(tuple_list)
    except ValueError:
        raise ValueError(
            "The input must a list of tuples \
                         with each tuple of length 2"
        )


def unlist(d: dict) -> dict:
    """Find all appearance of single element list in a dictionary and unlist it.

    :param: d: a python dictionary to be unlisted
    """
    if isinstance(d, list):
        if len(d) == 1:
            return d[0]
        return d
    if isinstance(d, dict):
        for key, val in d.items():
            if isinstance(val, list):
                if len(val) == 1:
                    d[key] = unlist(val[0])
            elif isinstance(val, dict):
                unlist(val)
        return d
    return d


def find_longest_common_path(paths) -> str:
    """Find longest common path among a list of paths.

    :param: paths: a list of paths, with '.' as the separator
    """
    return commonprefix(paths).rsplit(".", 1)[0]


def get_dict_values(py_dict: dict, excluded_keys: list = INTERNAL_KEYS) -> list:
    """Retrieve the values of a python dictionary.

    :param: py_dict: a python dictionary
    :param: excluded_keys: a list of keys to be excluded, \
        meaning the values of these keys should be excluded
    """
    return [v for k, v in py_dict.items() if k not in excluded_keys]


def get_primary_id_from_equivalent_ids(equivalent_ids: dict, _type: str):
    """Find primary id from equivalent id dict.

    :param: equivalent_ids: a dictionary containing all equivalent ids of a bio-entity
    :param: _type: the type of the bio-entity
    """
    if not equivalent_ids:
        return ""
    if _type in id_ranks:
        id_rank = id_ranks.get(_type)
        # loop through id_rank, return the first found id
        for _item in id_rank:
            if equivalent_ids.get(_item):
                return _item + ":" + str(equivalent_ids[_item][0])
    # if no id from id_rank found, return a random one from equivalent ids
    for k, v in equivalent_ids.items():
        if v:
            return k + ":" + str(v[0])
    return ""


def get_name_from_equivalent_ids(equivalent_ids, input_label=None):
    """Find name from equivalent id dict.

    :param: equivalent_ids: a dictionary containing all equivalent ids of a bio-entity.
    :param: input_label: desginated input_label
    """
    if input_label:
        return input_label
    if not equivalent_ids:
        return "unknown"
    if equivalent_ids.get("SYMBOL"):
        return equivalent_ids.get("SYMBOL")[0]
    if equivalent_ids.get("name"):
        return equivalent_ids.get("name")[0]
    for v in equivalent_ids.values():
        if v:
            if isinstance(v, list):
                return v[0]
            return v
    return "unknown"


def remove_prefix(_input, prefix):
    """Remove all prefixes in the input.
    
    :param: _input: the input
    :param: prefix: the prefix
    """
    if not prefix.endswith(":"):
        prefix += ":"
    if not _input:
        return _input
    if isinstance(_input, str):
        if _input.startswith(prefix):
            return _input[len(prefix) :]
        return _input
    if isinstance(_input, dict):
        new_result = {}
        for k, v in _input.items():
            if k.startswith(prefix):
                new_result[k[len(prefix) :]] = remove_prefix(v, prefix)
            else:
                new_result[k] = remove_prefix(v, prefix)
        return new_result
    if isinstance(_input, list):
        return [remove_prefix(item, prefix) for item in _input]
    return _input


def getPrefixFromCurie(curie: str) -> str:
    return curie.split(":")[0]


def getValFromCurie(curie: str) -> str:
    prefix = curie.split(":")[0]
    if prefix in ALWAYS_PREFIXED:
        return curie
    return curie.split(":", 1)[-1]


def getCurieFromVal(val: str, prefix: str) -> str:
    if prefix in ALWAYS_PREFIXED:
        return val
    return prefix + ":" + val


def dict_set2list(input_dict: dict) -> dict:
    return {k: list(v) for k, v in input_dict.items()}




from copy import deepcopy

from .utils import generateCurie, process_publications
from .json_transformer import Transformer


class BaseTransformer:
    def __init__(self, data):
        self.data = data
        self.edge = data.get("edge")

    def pairInputWithAPIResponse(self):
        """Create an object with key representing input, and value representing the output of API"""
        _input = generateCurie(self.edge["association"]["input_id"], self.edge["input"])
        return {_input: [self.data["response"]]}

    def wrap(self, res):
        """Wrapper functions to transform API response before passing to JSON Transformer"""
        if isinstance(res, list):
            res = {"data": res}
        return res

    def jsonTransform(self, res):
        """Transform Individual JSON response into Biolink compatible format
        :param: res: JSON response representing an output
        """
        return Transformer(res, self.edge["response_mapping"]).transform()

    def extractOutputIDs(self, res):
        """Retrieve all output IDs from response.
        :param: res: JSON response representing an output.
        """
        output_id_type = self.edge["association"]["output_id"]
        if output_id_type not in res:
            return []
        if isinstance(res[output_id_type], list):
            return [generateCurie(output_id_type, item) for item in res[output_id_type]]
        return [generateCurie(output_id_type, res[output_id_type])]

    def addEdgeInfo(self, _input, res):
        """Add edge information into individual output JSON.
        :param: res: JSON response representing an output.
        """
        if not res:
            return []
        output_ids = self.extractOutputIDs(res)
        result = []
        for item in output_ids:
            copy_item = deepcopy(item)
            copy_res = deepcopy(res)
            copy_res.update(
                {
                    "$reasoner_edge": self.edge.get("reasoner_edge", None),
                    "$association": self.edge.get("association", None),
                    "$input": _input,
                    "$output": copy_item,
                    "$original_input": self.edge.get("original_input", None),
                    "$input_resolved_identifiers": self.edge.get(
                        "input_resolved_identifiers", None
                    ),
                    "api": self.edge["association"]["api_name"],
                    "provided_by": res.get("source")
                    if res.get("source")
                    else [self.edge["association"].get("source", None)],
                }
            )
            copy_res = process_publications(copy_res)
            result.append(copy_res)
        return result

    def transform(self):
        """Main function to transform API response"""
        result = []
        responses = self.pairInputWithAPIResponse()
        for curie, res in responses.items():
            if isinstance(res, list) and len(res) > 0:
                for item in res:
                    item = self.wrap(item)
                    item = self.jsonTransform(item)
                    for val in item.values():
                        if isinstance(val, list) and len(val) > 0:
                            for v in val:
                                result += self.addEdgeInfo(curie, v)
                        else:
                            result += self.addEdgeInfo(curie, val)
        return result



from collections import defaultdict

from ..config_new import ID_RESOLVING_APIS
from ..exceptions.id_resolver import InvalidIDResolverInputError
from ..utils.common import getPrefixFromCurie


class Validator:
    def __init__(self, user_input):
        self.__user_input = user_input
        self.__valid = defaultdict(list)
        self.__invalid = defaultdict(list)

    def get_valid_inputs(self):
        return self.__valid

    def get_invalid_inputs(self):
        return self.__invalid

    def _validate_if_input_is_dict(self, user_input):
        if not isinstance(user_input, dict):
            raise InvalidIDResolverInputError(
                user_input,
                message="Your Input to ID Resolver is Invalid. It should be a dictionary!",
            )

    def _validate_if_values_of_input_is_list(self, user_input):
        for k, v in user_input.items():
            if not isinstance(v, list):
                raise InvalidIDResolverInputError(
                    user_input,
                    message="Your Input to ID Resolver is Invalid. All values of your input dictionary should be a list!",
                )

    def _validate_if_each_item_in_input_values_is_curie(self, user_input):
        for k, v in user_input.items():
            for _v in v:
                if not isinstance(_v, str) or ":" not in _v:
                    raise InvalidIDResolverInputError(
                        user_input,
                        message="Your Input to ID Resolver is Invalid. Each item in the values of your input dictionary should be a curie. Spotted {} is not a curie".format(
                            _v
                        ),
                    )

    def _check_if_semantic_type_can_be_resolved(self, user_input):
        res = {}
        for k, v in user_input.items():
            if k not in ID_RESOLVING_APIS:
                self.__invalid[k] = v
            else:
                res[k] = v
        return res

    def _check_if_prefix_can_be_resolved(self, user_input):
        for k, v in user_input.items():
            for _v in v:
                if getPrefixFromCurie(_v) not in ID_RESOLVING_APIS[k]["mapping"]:
                    self.__invalid[k].append(_v)
                else:
                    self.__valid[k].append(_v)

    def validate(self):
        self._validate_if_input_is_dict(self.__user_input)
        self._validate_if_values_of_input_is_list(self.__user_input)
        self._validate_if_each_item_in_input_values_is_curie(self.__user_input)
        tmp_valid_res = self._check_if_semantic_type_can_be_resolved(self.__user_input)
        self._check_if_prefix_can_be_resolved(tmp_valid_res)



from itertools import groupby

def restructure_reasoner_response(json_doc):
    """Restructure the API output from reasoner API.

    :param: json_doc: json output from reasoner API
    """
    edges = json_doc['knowledge_graph']['edges']
    if not edges:
        return {}
    res = {}
    edges = sorted(edges, key=lambda x: x['type'])
    for k, g in groupby(edges, lambda x: x['type']):
        res[k] = []
        for _item in g:
            if _item['target_id'].startswith("PANTHER.FAMILY"):
                _item['panther'] = _item['target_id'][15:]
            if _item['target_id'].startswith("CHEBI"):
                _item['chebi'] = _item['target_id']
            if _item['target_id'].startswith("CHEMBL:"):
                _item['chembl'] = _item['target_id'][7:]
            if _item['target_id'].startswith("MONDO:"):
                _item['mondo'] = _item['target_id'][6:]
            res[k].append(_item)
    return res


from collections import defaultdict

from ..config_new import ID_RESOLVING_APIS, ALWAYS_PREFIXED


class BioThingsParser:
    def __init__(self, response, semanticType, prefix):
        self.response = response
        self.semanticType = semanticType
        self.prefix = prefix
        self.invalid = []

    def parse(self):
        result = {}
        if not self.response or len(self.response) == 0:
            return result
        mapping = ID_RESOLVING_APIS[self.semanticType]["mapping"]
        for rec in self.response:
            if not rec:
                continue
            if self.prefix in ALWAYS_PREFIXED:
                curie = rec.get("query")
            else:
                curie = self.prefix + ":" + rec.get("query")
            if "notfound" in rec:
                self.invalid.append(curie)
                continue
            if curie not in result:
                result[curie] = defaultdict(set)
            if (
                self.semanticType == "Gene"
                and "symbol" in rec
                and "label" not in result[curie]
            ):
                result[curie]["label"] = (
                    set([rec["symbol"]])
                    if not isinstance(rec["symbol"], list)
                    else set([rec["symbol"][0]])
                )
            for prefix, fields in mapping.items():
                for field in fields:
                    if field in rec:
                        if isinstance(rec[field], list):
                            if prefix == "name" and "label" not in result[curie]:
                                result[curie]["label"].add(str(rec[field][0]))
                            rec[field] = [str(item) for item in rec[field]]
                            result[curie][prefix].update(rec[field])
                        else:
                            if prefix == "name" and "label" not in result[curie]:
                                result[curie]["label"].add(str(rec[field]))
                            result[curie][prefix].add(str(rec[field]))
        return self.restructureOutput(result, self.semanticType)

    def restructureOutput(self, res, semanticType):
        result = {}
        for curie in self.invalid:
            db_id = curie
            if curie.split(":")[0] not in ALWAYS_PREFIXED:
                db_id = curie.split(":", 1)[-1]
            result[curie] = {
                "id": {"identifier": curie, "label": curie},
                "curies": [curie],
                "db_ids": {curie.split(":")[0]: [db_id]},
                "type": semanticType,
                "flag": "failed",
            }
        ranks = ID_RESOLVING_APIS[semanticType]["id_ranks"]
        for curie, resolved_ids in res.items():
            ids = set()
            result[curie] = {"id": {}, "db_ids": {}, "type": semanticType}
            if "label" in resolved_ids:
                result[curie]["id"]["label"] = list(res[curie]["label"])[0]
            elif "name" in resolved_ids:
                result[curie]["id"]["label"] = list(res[curie]["name"])[0]
            else:
                result[curie]["id"]["label"] = curie
            primary_id_found = False
            for _id in ranks:
                if resolved_ids[_id] and len(resolved_ids[_id]) > 0:
                    result[curie]["db_ids"][_id] = list(res[curie][_id])
                    for item in res[curie][_id]:
                        item_curie = item
                        if _id not in ALWAYS_PREFIXED:
                            item_curie = _id + ":" + str(item)
                        if not primary_id_found:
                            result[curie]["id"]["identifier"] = item_curie
                            primary_id_found = True
                        if _id != "name":
                            ids.add(item_curie)
            result[curie]["curies"] = list(ids)
        return result



def restructure_stanford_response(json_doc):
    """Restructure the JSON output from stanford KP.

    :param: json_doc: the API response from stanford KP
    """
    # convert from list to dict
    if isinstance(json_doc, list):
        json_doc = {'data': json_doc}
    if 'data' in json_doc:
        for _doc in json_doc['data']:
            # store restructured attributes info
            new_attr = {}
            for _attr in _doc['attributes']:
                if 'attributeValueTermUri' not in _attr:
                    new_attr[_attr['attributeName'].replace(' ', '_')] = _attr['attributeValue']
                else:
                    new_attr[_attr['attributeName'].replace(' ', '_')] = {'name': _attr['attributeValue']}
                    if 'PATO_' in _attr['attributeValueTermUri']:
                        new_attr[_attr['attributeName'].replace(' ', '_')]['pato'] = 'PATO:' + _attr['attributeValueTermUri'].split('_')[-1]
                    elif 'MONDO_' in _attr['attributeValueTermUri']:
                        new_attr[_attr['attributeName'].replace(' ', '_')]['mondo'] = 'MONDO:' + _attr['attributeValueTermUri'].split('_')[-1]
                    elif 'BTO_' in _attr['attributeValueTermUri']:
                        new_attr[_attr['attributeName'].replace(' ', '_')]['bto'] = 'BTO:' + _attr['attributeValueTermUri'].split('_')[-1]
                    elif 'CLO_' in _attr['attributeValueTermUri']:
                        new_attr[_attr['attributeName'].replace(' ', '_')]['clo'] = 'CLO:' + _attr['attributeValueTermUri'].split('_')[-1]
            _doc['attributes'] = new_attr
        return json_doc
    return {}



from collections import defaultdict

from ..config_new import ID_RESOLVING_APIS
from ..utils.common import getPrefixFromCurie, getValFromCurie


class CurieGroup:
    def __init__(self, semanticType, curies):
        self.semanticType = semanticType
        self.curies = curies

    @staticmethod
    def _findAPI(semanticType):
        return ID_RESOLVING_APIS.get(semanticType, {})

    def groupCuriesByPrefix(self, curies: list):
        grped = defaultdict(set)
        for curie in curies:
            prefix = getPrefixFromCurie(curie)
            val = getValFromCurie(curie)
            grped[prefix].add(val)
        return grped
